# Source: https://github.com/bobmyhill/mdx_bib/blob/6b13bbbc407617a5e93ed0f8a0e5e4c52f73f677/mdx_bib.py
import re
import string
from collections import Counter, OrderedDict
from xml.etree import ElementTree as etree

from markdown.extensions import Extension
from markdown.inlinepatterns import Pattern
from markdown.preprocessors import Preprocessor
from markdown.treeprocessors import Treeprocessor
from pybtex.database.input import bibtex
from pybtex.exceptions import PybtexError

BRACKET_RE = re.compile(r"\[([^\[]+)\]")
CITE_RE = re.compile(r"@(\w+)")
DEF_RE = re.compile(r"\A {0,3}\[@(\w+)\]:\s*(.*)")
INDENT_RE = re.compile(r"\A\t| {4}(.*)")

CITATION_RE = r"@(\w+)"


class Bibliography(object):
    """Keep track of document references and citations for exporting."""

    def __init__(self, extension, bibtex_file, order):
        self.extension = extension
        self.order = order

        self.citations = OrderedDict()
        self.references = {}

        if bibtex_file:
            try:
                parser = bibtex.Parser()
                self.bibsource = parser.parse_file(bibtex_file).entries
                self.labels = {
                    id: self.formatCitation(self.bibsource[id]) for id in self.bibsource
                }
                for value, occurrences in Counter(self.labels.values()).items():
                    if occurrences > 1:
                        for xkey, xvalue in self.labels.items():
                            i = 0
                            if xvalue == value:
                                self.labels[xkey] = (
                                    f"{xvalue}{string.ascii_lowercase[i]}"
                                )
                                i += 1

            except PybtexError:
                print("Error loading bibtex file")
                self.bibsource = {}
                self.labels = {}
        else:
            self.bibsource = {}

    def addCitation(self, citekey):
        self.citations[citekey] = self.citations.get(citekey, 0) + 1

    def setReference(self, citekey, reference):
        self.references[citekey] = reference

    def citationID(self, citekey):
        return "cite-" + citekey

    def referenceID(self, citekey):
        return "ref-" + citekey

    def formatAuthor(self, author):
        out = f"{author.last_names[0]} {author.first_names[0][0]}."
        if author.middle_names:
            out += f"{author.middle_names[0][0]}."
        return out.replace("{", "").replace("}", "")

    def formatAuthorSurname(self, author):
        out = author.last_names[0]
        return out.replace("{", "").replace("}", "")

    def formatReference(self, ref):
        author_list = list(map(self.formatAuthor, ref.persons["author"]))

        if len(author_list) == 1:
            authors = author_list[0]
        else:
            authors = ", ".join(author_list[:-1])
            authors += f" and {author_list[-1]}"

        # Harvard style
        # Surname, Initial, ... and Last_Surname,
        # Initial, Year. Title. Journal, Volume(Issue), pages. doi.

        title = ref.fields["title"].replace("{", "").replace("}", "")
        journal = ref.fields.get("journal", "")
        volume = ref.fields.get("volume", "")
        issue = ref.fields.get("issue", "")
        year = ref.fields.get("year")
        pages = ref.fields.get("pages")
        doi = ref.fields.get("doi")

        reference = f"<p>{authors}, {year}. {title}."
        if journal:
            reference += f" <i>{journal}</i>."
            if volume:
                reference += f" <i>{volume}</i>"
            if issue:
                reference += f"({issue})"
            if pages:
                reference += f", pp.{pages}"
            reference += "."
        if doi:
            reference += (
                f' <a href="https://dx.doi.org/{doi}" target="_blank">{doi}</a>'
            )
        reference += "</p>"

        return reference

    def formatCitation(self, ref):
        author_list = list(map(self.formatAuthorSurname, ref.persons["author"]))
        year = ref.fields.get("year")

        if len(author_list) == 1:
            citation = f"{author_list[0]}"
        elif len(author_list) == 2:
            citation = f"{author_list[0]} and {author_list[1]}"
        else:
            citation = f"{author_list[0]} et al."

        citation += f", {year}"

        return citation

    def makeBibliography(self, root):
        if self.order == "alphabetical":
            raise (NotImplementedError)

        div = etree.Element("div")
        div.set("class", "references")

        if not self.citations:
            return div

        table = etree.SubElement(div, "table")
        tbody = etree.SubElement(table, "tbody")
        for id in self.citations:
            tr = etree.SubElement(tbody, "tr")
            tr.set("id", self.referenceID(id))
            ref_id = etree.SubElement(tr, "td")
            ref_txt = etree.SubElement(tr, "td")
            if id in self.references:
                self.extension.parser.parseChunk(ref_txt, self.references[id])
                ref_id.text = self.labels[id]
            elif id in self.bibsource:
                ref_txt.text = self.formatReference(self.bibsource[id])
                ref_id.text = self.labels[id]
            else:
                ref_txt.text = "Missing citation"

        return div

    def clearCitations(self):
        self.citations = OrderedDict()


class CitationsPreprocessor(Preprocessor):
    """Gather reference definitions and citation keys."""

    def __init__(self, bibliography):
        self.bib = bibliography

    def subsequentIndents(self, lines, i):
        """Concatenate consecutive indented lines."""
        linesOut = []
        while i < len(lines):
            m = INDENT_RE.match(lines[i])
            if m:
                linesOut.append(m.group(1))
                i += 1
            else:
                break
        return " ".join(linesOut), i

    def run(self, lines):
        linesOut = []
        i = 0

        while i < len(lines):
            # Check to see if the line starts a reference definition
            m = DEF_RE.match(lines[i])
            if m:
                key = m.group(1)
                reference = m.group(2)
                indents, i = self.subsequentIndents(lines, i + 1)
                reference += " " + indents

                self.bib.setReference(key, reference)
                continue

            # Look for all @citekey patterns inside hard brackets
            for bracket in BRACKET_RE.findall(lines[i]):
                for c in CITE_RE.findall(bracket):
                    self.bib.addCitation(c)
            linesOut.append(lines[i])
            i += 1

        return linesOut


class CitationsPattern(Pattern):
    """Handles converting citations keys into links."""

    def __init__(self, pattern, bibliography):
        super(CitationsPattern, self).__init__(pattern)
        self.bib = bibliography

    def handleMatch(self, m):
        id = m.group(2)
        if id in self.bib.bibsource:
            a = etree.Element("a")
            a.set("id", self.bib.citationID(id))
            a.set("href", "#" + self.bib.referenceID(id))
            a.set("class", "citation")
            a.text = self.bib.labels[id]
            return a
        else:
            return None


class CitationsTreeprocessor(Treeprocessor):
    """Add a bibliography/reference section to the end of the document."""

    def __init__(self, bibliography):
        self.bib = bibliography

    def run(self, root):
        citations = self.bib.makeBibliography(root)
        root.append(citations)
        self.bib.clearCitations()


class CitationsExtension(Extension):
    def __init__(self, *args, **kwargs):
        self.config = {
            "bibtex_file": ["", "Bibtex file path"],
            "order": ["unsorted", "Order of the references (unsorted, alphabetical)"],
        }
        super(CitationsExtension, self).__init__(*args, **kwargs)
        self.bib = Bibliography(
            self,
            self.getConfig("bibtex_file"),
            self.getConfig("order"),
        )

    def extendMarkdown(self, md):
        md.registerExtension(self)
        self.parser = md.parser
        self.md = md

        md.preprocessors.register(CitationsPreprocessor(self.bib), "mdx_bib", 15)
        md.inlinePatterns.register(
            CitationsPattern(CITATION_RE, self.bib), "mdx_bib", 175
        )
        md.treeprocessors.register(CitationsTreeprocessor(self.bib), "mdx_bib", 25)


def makeExtension(*args, **kwargs):
    return CitationsExtension(*args, **kwargs)



import isce3
import numpy as np
from numpy.typing import ArrayLike
from opera_utils import (
    get_cslc_orbit,
    get_lonlat_grid,
    get_radar_wavelength,
)

from dolphin._types import Filename


def compute(
    llh: ArrayLike,
    ref_pos: ArrayLike,
    sec_pos: ArrayLike,
    ref_range: float,
    sec_range: float,
    ref_vel: ArrayLike,
    ell: isce3.core.Ellipsoid,
):
    """Compute the perpendicular baseline at a given geographic position.

    Parameters
    ----------
    llh : ArrayLike
        Lon/Lat/Height vector specifying the target position.
        Lon and Lat must be in radians, not degrees.
    ref_pos : ArrayLike
        Reference position vector (x, y, z) in ECEF coordinates.
    sec_pos : ArrayLike
        Secondary position vector (x, y, z) in ECEF coordinates.
    ref_range : float
        Range from the reference satellite to the target.
    sec_range : float
        Range from the secondary satellite to the target.
    ref_vel : ArrayLike
        Velocity vector (vx, vy, vz) of the reference satellite in ECEF coordinates.
    ell : isce3.core.Ellipsoid
        Ellipsoid for the target surface.

    Returns
    -------
    float
        Perpendicular baseline, in meters.

    """
    # Difference in position between the two passes
    baseline = np.linalg.norm(sec_pos - ref_pos)

    # Compute angle between LOS vector and baseline vector
    # via the law of cosines
    cos_theta = (ref_range**2 + baseline**2 - sec_range**2) / (2 * ref_range * baseline)

    sin_theta = np.sqrt(1 - cos_theta**2)
    perp = baseline * sin_theta
    # parallel_baseline = baseline * cosine_theta

    target_xyz = ell.lon_lat_to_xyz(llh)
    direction = np.sign(
        np.dot(np.cross(target_xyz - ref_pos, sec_pos - ref_pos), ref_vel)
    )

    return direction * perp


def compute_baselines(
    h5file_ref: Filename,
    h5file_sec: Filename,
    height: float = 0.0,
    latlon_subsample: int = 100,
    threshold: float = 1e-08,
    maxiter: int = 50,
    delta_range: float = 10.0,
):
    """Compute the perpendicular baseline at a subsampled grid for two CSLCs.

    Parameters.
    ----------
    h5file_ref : Filename
        Path to reference OPERA S1 CSLC HDF5 file.
    h5file_sec : Filename
        Path to secondary OPERA S1 CSLC HDF5 file.
    height: float
        Target height to use for baseline computation.
        Default = 0.0
    latlon_subsample: int
        Factor by which to subsample the CSLC latitude/longitude grids.
        Default = 30
    threshold : float
        isce3 geo2rdr: azimuth time convergence threshold in meters
        Default = 1e-8
    maxiter : int
        isce3 geo2rdr: Maximum number of Newton-Raphson iterations
        Default = 50
    delta_range : float
        isce3 geo2rdr: Step size used for computing derivative of doppler
        Default = 10.0

    Returns
    -------
    lon : np.ndarray
        2D array of longitude coordinates in degrees.
    lat : np.ndarray
        2D array of latitude coordinates in degrees.
    baselines : np.ndarray
        2D array of perpendicular baselines

    """
    lon_grid, lat_grid = get_lonlat_grid(h5file_ref, subsample=latlon_subsample)
    lon_arr = lon_grid.ravel()
    lat_arr = lat_grid.ravel()

    ellipsoid = isce3.core.Ellipsoid()
    zero_doppler = isce3.core.LUT2d()
    wavelength = get_radar_wavelength(h5file_ref)
    side = isce3.core.LookSide.Right

    orbit_ref = get_cslc_orbit(h5file_ref)
    orbit_sec = get_cslc_orbit(h5file_sec)

    baselines = []
    for lon, lat in zip(lon_arr, lat_arr):
        llh_rad = np.deg2rad([lon, lat, height]).reshape((3, 1))
        az_time_ref, range_ref = isce3.geometry.geo2rdr(
            llh_rad,
            ellipsoid,
            orbit_ref,
            zero_doppler,
            wavelength,
            side,
            threshold=threshold,
            maxiter=maxiter,
            delta_range=delta_range,
        )
        az_time_sec, range_sec = isce3.geometry.geo2rdr(
            llh_rad,
            ellipsoid,
            orbit_sec,
            zero_doppler,
            wavelength,
            side,
            threshold=threshold,
            maxiter=maxiter,
            delta_range=delta_range,
        )

        pos_ref, velocity = orbit_ref.interpolate(az_time_ref)
        pos_sec, _ = orbit_sec.interpolate(az_time_sec)
        b = compute(
            llh_rad, pos_ref, pos_sec, range_ref, range_sec, velocity, ellipsoid
        )

        baselines.append(b)

    baseline_grid = np.array(baselines).reshape(lon_grid.shape)
    return lon_grid, lat_grid, baseline_grid



from __future__ import annotations

import logging
from math import exp, sqrt
from typing import Optional

import numba
import numpy as np
from numba import cuda
from numpy.typing import ArrayLike

from dolphin._types import Strides
from dolphin.utils import _get_slices, compute_out_shape

from ._common import remove_unconnected

logger = logging.getLogger(__name__)


_get_slices = numba.njit(_get_slices)


def estimate_neighbors(
    amp_stack: ArrayLike,
    halfwin_rowcol: tuple[int, int],
    alpha: float,
    strides: Optional[dict[str, int]] = None,
    is_sorted: bool = False,
    prune_disconnected: bool = False,
):
    """Estimate the  at all pixels of `slc_stack` on the GPU."""
    # estimate_neighbors_gpu(
    #     sorted_amp_stack,
    #     halfwin_rowcol,
    #     strides_rowcol,
    #     alpha,
    #     neighbor_arrays,
    # )

    if strides is None:
        strides = {"x": 1, "y": 1}
    sorted_amp_stack = amp_stack if is_sorted else np.sort(amp_stack, axis=0)

    num_slc, rows, cols = sorted_amp_stack.shape
    ecdf_dist_cutoff = _get_ecdf_critical_distance(num_slc, alpha)
    logger.debug(f"ecdf_dist_cutoff: {ecdf_dist_cutoff}")

    strides_rowcol = strides["y"], strides["x"]
    out_rows, out_cols = compute_out_shape((rows, cols), Strides(*strides_rowcol))
    half_row, half_col = halfwin_rowcol
    is_shp = np.zeros(
        (out_rows, out_cols, 2 * half_row + 1, 2 * half_col + 1), dtype=np.bool_
    )

    _loop_over_neighbors(
        sorted_amp_stack,
        halfwin_rowcol,
        strides_rowcol,
        ecdf_dist_cutoff,
        prune_disconnected,
        is_shp,
    )

    return is_shp


@numba.njit(parallel=True, nogil=True)
def _loop_over_neighbors(
    sorted_amp_stack,
    halfwin_rowcol: tuple[int, int],
    strides_rowcol: tuple[int, int],
    ecdf_dist_cutoff: float,
    prune_disconnected: bool,
    is_shp: ArrayLike,
):
    """Estimate the SHPs of each pixel of `amp_stack` on the CPU."""
    num_slc, in_rows, in_cols = sorted_amp_stack.shape
    out_rows, out_cols = is_shp.shape[:2]

    half_row, half_col = halfwin_rowcol
    row_strides, col_strides = strides_rowcol
    r_start = row_strides // 2
    c_start = col_strides // 2
    # location to start counting from in the larger input
    r0, c0 = row_strides // 2, col_strides // 2

    for out_r in numba.prange(out_rows):
        for out_c in range(out_cols):
            in_r = r0 + out_r * row_strides
            in_c = c0 + out_c * col_strides
            # Clamp the window to the image bounds
            (r_start, r_end), (c_start, c_end) = _get_slices(
                half_row, half_col, in_r, in_c, in_rows, in_cols
            )

            amp_block = sorted_amp_stack[:, r_start:r_end, c_start:c_end]
            neighbors = is_shp[out_r, out_c, :, :]
            _set_neighbors(amp_block, halfwin_rowcol, ecdf_dist_cutoff, neighbors)
            if prune_disconnected:
                remove_unconnected(is_shp[out_r, out_c], inplace=True)

    return is_shp


# GPU version of the SHP finding algorithm using KS test
@cuda.jit
def estimate_neighbors_gpu(
    sorted_amp_stack,
    halfwin_rowcol: tuple[int, int],
    strides_rowcol: tuple[int, int],
    alpha: float,
    is_shp: ArrayLike,
):
    """Estimate the SHPs of each pixel of `amp_stack` on the GPU."""
    # Get the global position within the 2D GPU grid
    out_x, out_y = cuda.grid(2)
    out_rows, out_cols = is_shp.shape[:2]
    # Check if we are within the bounds of the array
    if out_y >= out_rows or out_x >= out_cols:
        return

    num_slc, rows, cols = sorted_amp_stack.shape
    ecdf_dist_cutoff = _get_ecdf_critical_distance(num_slc, alpha)
    half_row, half_col = halfwin_rowcol

    row_strides, col_strides = strides_rowcol
    r_start = row_strides // 2
    c_start = col_strides // 2
    in_r = r_start + out_y * row_strides
    in_c = c_start + out_x * col_strides

    # Get the input slices, clamping the window to the image bounds
    (r_start, r_end), (c_start, c_end) = _get_slices(
        half_row, half_col, in_r, in_c, rows, cols
    )

    amp_block = sorted_amp_stack[:, r_start:r_end, c_start:c_end]
    neighbors_pixel = is_shp[out_y, out_x, :, :]
    _set_neighbors(amp_block, halfwin_rowcol, ecdf_dist_cutoff, neighbors_pixel)
    remove_unconnected(is_shp[out_y, out_x], inplace=True)


@numba.njit(nogil=True)
def _set_neighbors(amp_block, halfwin_rowcol, ecdf_dist_cutoff, neighbors):
    _, rows, cols = amp_block.shape

    if rows < 2 * halfwin_rowcol[0] + 1 or cols < 2 * halfwin_rowcol[1] + 1:
        # not enough neighbors to test, make all false
        return

    # get the center pixel
    r_c, c_c = rows // 2, cols // 2
    x1 = amp_block[:, r_c, c_c]
    # TODO: if this is a bottleneck, we can do something smarter
    # by computing only the bottom right corner, then mirroring
    for i in range(rows):
        for j in range(cols):
            if i == r_c and j == c_c:
                neighbors[i, j] = False
                continue
            x2 = amp_block[:, i, j]

            ecdf_max_dist = _get_max_cdf_dist(x1, x2)

            neighbors[i, j] = ecdf_max_dist < ecdf_dist_cutoff


@numba.njit(nogil=True)
def _get_max_cdf_dist(x1, x2):
    """Get the maximum CDF distance between two arrays.

    Parameters
    ----------
    x1 : np.ndarray
        First array, size n, sorted
    x2 : np.ndarray
        Second array, size n, sorted

    Returns
    -------
    float
        Maximum empirical CDF distance between the two arrays

    Examples
    --------
    >>> x1 = np.array([1, 2, 3, 4, 5])
    >>> x2 = np.array([1, 2, 3, 4, 5])
    >>> print(f"{_get_max_cdf_dist(x1, x2):.1f}")  # doctest: +NUMBER
    0.0
    >>> x2 = np.array([2, 3, 4, 5, 6])
    >>> round(_get_max_cdf_dist(x1, x2), 2)  # doctest: +NUMBER
    0.2
    >>> round(_get_max_cdf_dist(x2, x1), 2)  # doctest: +NUMBER
    0.2
    >>> x2 = np.array([6, 7, 8, 9, 10])
    >>> _get_max_cdf_dist(x1, x2)  # doctest: +NUMBER
    1.0

    """
    n = x1.shape[0]
    i1 = i2 = i_out = 0
    cdf1 = cdf2 = 0
    max_dist = 0
    while i_out < 2 * n:
        if i1 == n:
            cdf2 += 1 / n
            i2 += 1
        elif i2 == n or (x1[i1] < x2[i2]):
            cdf1 += 1 / n
            i1 += 1
        elif x1[i1] > x2[i2]:
            cdf2 += 1 / n
            i2 += 1
        else:  # a tie
            cdf1 += 1 / n
            cdf2 += 1 / n
            i1 += 1
            i2 += 1
            i_out += 1  # jumping 2 ahead for tie
        i_out += 1
        max_dist = max(max_dist, abs(cdf1 - cdf2))
    return max_dist


@numba.njit(nogil=True)
def _get_ecdf_critical_distance(nslc, alpha):
    N = nslc / 2.0
    cur_dist = 0.01
    critical_distance = 0.1
    sqrt_N = sqrt(N)
    while cur_dist <= 1.0:
        value = cur_dist * (sqrt_N + 0.12 + 0.11 / sqrt_N)
        pvalue = 0
        for t in range(1, 101):
            pvalue += ((-1) ** (t - 1)) * exp(-2 * (value**2) * (t**2))
        pvalue = max(0.0, min(1.0, 2 * pvalue))

        if pvalue <= alpha:
            critical_distance = cur_dist
            break

        cur_dist += 0.001
    return critical_distance



from __future__ import annotations

import argparse

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import patches

from dolphin import shp
from dolphin.io import VRTStack


def plot_shps(
    slc_stack,
    mean,
    var,
    half_window: dict[str, int],
    display_arr=None,
    shp_methods=None,
    shp_alpha=0.05,
    shp_nslc=None,
    cmap_nmap="Reds_r",
    block=False,
):
    """Interactively plot the neighborhood map over an SLC image.

    Click on a pixel to see the neighborhood map.
    """
    if shp_methods is None:
        shp_methods = ["glrt", "ks"]
    if isinstance(slc_stack, str):
        slc_stack = VRTStack(slc_stack, write_file=False)
    if display_arr is None:
        display_arr = mean
    if shp_nslc is None:
        shp_nslc = slc_stack.shape[0]
    if not isinstance(shp_alpha, list):
        shp_alpha = [shp_alpha] * len(shp_methods)

    ny, nx = half_window["y"], half_window["x"]
    halfwin_rowcol = (ny, nx)
    # window = (1 + 2 * ny, 1 + 2 * nx)

    fig, axes = plt.subplots(
        ncols=len(shp_methods), sharex=True, sharey=True, squeeze=False
    )
    axes = axes.ravel()
    for ax, shp_method in zip(axes, shp_methods):
        ax.imshow(_scale_mag(mean), cmap="gray")
        ax.set_title(shp_method)

    def onclick(event):
        # Ignore right/middle click, clicks off image
        if event.button != 1 or not event.inaxes:
            return
        if event.inaxes not in axes:
            return
        state = fig.canvas.manager.toolbar.mode
        if state != "":  # Zoom/other tool is active
            return

        # Save limits to restore after adding neighborhoods
        xlim = axes[0].get_xlim()
        ylim = axes[0].get_ylim()
        row, col = int(event.ydata), int(event.xdata)
        # Somehow clicked outside image, but in axis
        if row >= display_arr.shape[0] or col >= display_arr.shape[1]:
            return

        for i, shp_method in enumerate(shp_methods):
            # Calc SHPS:
            ax = axes[i]
            rs, cs = _get_slices(row, col, ny, nx)
            amp_stack = np.abs(slc_stack[:, rs, cs]) if shp_method == "ks" else None

            cur_shps = shp.estimate_neighbors(
                halfwin_rowcol=halfwin_rowcol,
                alpha=shp_alpha[i],
                mean=mean[rs, cs],
                var=var[rs, cs],
                nslc=shp_nslc,
                amp_stack=amp_stack,
                method=shp_method,
            )
            cur_shps = cur_shps[ny, nx, :, :]
            n_img = np.ma.masked_not_equal(cur_shps, True)
            extent = _get_extent(row, col, ny, nx)

            # Remove old neighborhood images
            if len(ax.get_images()) > 1:
                # event.inaxes.get_images()[1].remove()
                ax.get_images()[-1].remove()

            ax.imshow(
                n_img,
                cmap=cmap_nmap,
                alpha=0.8,
                extent=extent,
                origin="lower",
            )
            # Remove old neighborhood patches
            for p in ax.patches:
                p.remove()
            # add a rectangle around the neighborhood
            rect = patches.Rectangle(
                (extent[0], extent[2]),
                1 + 2 * nx,
                1 + 2 * ny,
                linewidth=1,
                edgecolor="r",
                facecolor="none",
            )
            ax.add_patch(rect)

            # Restore original viewing bounds
            ax.set_xlim(xlim)
            ax.set_ylim(ylim)

            fig.canvas.draw()

    fig.canvas.mpl_connect("button_press_event", onclick)
    fig.tight_layout()

    plt.show(block=block)


def _get_extent(row, col, ny, nx):
    """Get the row/col extent of the window surrounding a pixel."""
    # Matplotlib extent is (left, right, bottom, top)
    # Also the extent for normal `imshow` is shifted by -0.5
    return col - nx - 0.5, col + nx + 1 - 0.5, row - ny - 0.5, row + ny + 1 - 0.5


def _get_slices(row, col, ny, nx):
    """Get the row/col slice of the window surrounding a pixel."""
    return slice(row - ny, row + ny + 1), slice(col - nx, col + nx + 1)


def _scale_mag(img, exponent=0.3, max_pct=99.95):
    """Scale the magnitude of complex radar image for better display."""
    out = np.abs(img) ** exponent
    max_val = np.nanpercentile(out, max_pct)
    return np.clip(out, None, max_val)


def get_cli_args():
    """Get command line arguments."""
    parser = argparse.ArgumentParser(
        description="Interactively view neighborhood maps over an SLC."
    )
    parser.add_argument(
        "-n",
        "--nmap-file",
        help="Neighborhood map file",
        required=True,
    )
    parser.add_argument(
        "--slc-stack-file",
        help="SLC stack file",
    )
    parser.add_argument(
        "--slc-stack-bands",
        nargs="+",
        type=int,
        default=[1, 2, 3, 4, 5],
        help="Bands to use from SLC stack file",
    )
    parser.add_argument(
        "--slc-file",
        help="Alternative background: a single SLC filename.",
    )
    parser.add_argument(
        "--ps-file",
        default="PS/ps_pixels",
        help="PS file to overlay",
    )
    parser.add_argument(
        "--cmap-nmap",
        default="Reds_r",
        help="Colormap for neighborhood map",
    )
    parser.add_argument(
        "--shp-alpha",
        type=float,
        default=0.01,
        help="Alpha significance value for neighborhood map",
    )
    parser.add_argument(
        "--no-block",
        action="store_true",
        help="Don't block the matplotlib window (default is blocking)",
    )
    return parser.parse_args()


def run_cli():
    """Run the command line interface."""
    args = get_cli_args()
    plot_shps(
        nmap_filename=args.nmap_file,
        slc_stack_filename=args.slc_stack_file,
        slc_stack_bands=args.slc_stack_bands,
        slc_filename=args.slc_file,
        ps_filename=args.ps_file,
        cmap_nmap=args.cmap_nmap,
        block=not args.no_block,
    )


if __name__ == "__main__":
    run_cli()



from pathlib import Path

import numpy as np
import numpy.testing as npt
import pytest

from dolphin._types import Strides
from dolphin.phase_link import simulate
from dolphin.phase_link._ps_filling import fill_ps_pixels, get_max_idxs

RNG = np.random.default_rng()


@pytest.mark.parametrize("use_max_ps", [True, False])
@pytest.mark.parametrize("strides", [1, 2, 3, 4])
@pytest.mark.parametrize("rows", list(range(10, 16)))
@pytest.mark.parametrize("cols", list(range(10, 12)))
def test_ps_fill(C_truth, use_max_ps, strides, rows, cols):
    """Test substituting the PS pixel phases into the phase linking result."""
    num_acq = 9
    C = C_truth[0][:num_acq, :num_acq]

    slc_stack = simulate.simulate_neighborhood_stack(C, rows * cols).reshape(
        num_acq, rows, cols
    )

    # Fake an all zero-phase solution that will get filled with the PS pixel phases
    pl_est = np.ones((num_acq, rows // strides, cols // strides), dtype=np.complex64)
    temp_coh = np.zeros(pl_est.shape[1:])

    ps_idx = 2
    ps_mask = np.zeros((rows, cols), dtype=bool)
    ps_mask[ps_idx, ps_idx] = True

    fill_ps_pixels(
        pl_est,
        temp_coh,
        slc_stack,
        ps_mask,
        Strides(strides, strides),
        None,  # avg_mag
        use_max_ps=use_max_ps,
    )

    ps_phase = slc_stack[:, ps_idx, ps_idx]
    ps_phase *= ps_phase[0].conj()  # Reference to first acquisition

    out_idx = ps_idx // strides
    npt.assert_array_almost_equal(
        np.angle(ps_phase), np.angle(pl_est[:, out_idx, out_idx])
    )

    assert temp_coh[out_idx, out_idx] == 1


def test_get_max_idxs_no_look():
    arr = np.array([[5, 1, 3], [4, 9, 2], [8, 7, 6]])
    row_looks = 1
    col_looks = 1
    expected_rows, expected_cols = np.where(arr == arr)
    actual_rows, actual_cols = get_max_idxs(arr, row_looks, col_looks)
    assert np.array_equal(actual_rows, expected_rows)
    assert np.array_equal(actual_cols, expected_cols)


def test_get_max_idxs_2x2_window():
    arr = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]])
    row_looks = 2
    col_looks = 2
    expected_rows = np.array([1, 1, 3, 3])
    expected_cols = np.array([1, 3, 1, 3])
    actual_rows, actual_cols = get_max_idxs(arr, row_looks, col_looks)
    assert np.array_equal(actual_rows, expected_rows)
    assert np.array_equal(actual_cols, expected_cols)


def test_get_max_idxs_3x3_mixed_values():
    arr = np.array([[1, 5, 3], [4, 20, 2], [8, 0, 6]])
    row_looks = 2
    col_looks = 2
    expected_rows = np.array([1])
    expected_cols = np.array([1])
    actual_rows, actual_cols = get_max_idxs(arr, row_looks, col_looks)
    assert np.array_equal(actual_rows, expected_rows)
    assert np.array_equal(actual_cols, expected_cols)


def generate_manual_expected_indices(n, looks):
    # For the simple reshaping of arange to a rectangle, the max values are
    # the bottom right corners of each window.
    step = looks
    num_windows = n // looks
    end_idx = looks - 1
    rows = np.repeat(np.arange(end_idx, step * num_windows, step), num_windows)
    cols = np.tile(np.arange(end_idx, step * num_windows, step), num_windows)
    return rows, cols


@pytest.mark.parametrize("n", list(range(10, 15)))
@pytest.mark.parametrize("looks", [2, 3])
def test_get_max_idxs_param(n, looks):
    arr = np.arange(n * n).reshape(n, n)
    expected = generate_manual_expected_indices(n, looks)
    actual_rows, actual_cols = get_max_idxs(arr, looks, looks)
    assert np.array_equal(actual_rows, expected[0])
    assert np.array_equal(actual_cols, expected[1])


def test_get_max_idxs_uneven_shape():
    arr = np.arange(9 * 11).reshape(9, 11)
    rows, cols = get_max_idxs(arr, 2, 3)
    expected = np.array([[13, 16, 19], [35, 38, 41], [57, 60, 63], [79, 82, 85]])
    npt.assert_array_equal(arr[rows, cols], expected.ravel())


def test_failing_ps_size():
    data = np.load(Path(__file__).parent / "data/ps-fix/failing_data_idxs.npz")
    shape2d = (522, 534)
    slc_stack = RNG.normal(size=(15, *shape2d)) + 1j * RNG.normal(size=(15, *shape2d))
    slc_stack = slc_stack.astype("complex64")
    slc_stack[data["slc_nan_t"], data["slc_nan_y"], data["slc_nan_x"]] = np.nan
    reference_idx = 0

    mags = RNG.normal(size=shape2d) ** 2
    avg_mag = np.full_like(mags, np.nan)
    rows, cols = data["not_nan_rows"], data["not_nan_cols"]
    avg_mag[rows, cols] = mags[rows, cols]
    ps_mask = np.zeros(shape2d, dtype=bool)
    ps_mask[rows, cols] = True

    small_shape2d = (174, 89)
    temp_coh = RNG.random(size=small_shape2d)
    strides = Strides(3, 6)
    cpx_phase = RNG.normal(size=(15, *small_shape2d)) + 1j * RNG.normal(
        size=(15, *small_shape2d)
    )

    fill_ps_pixels(
        cpx_phase=cpx_phase,
        temp_coh=temp_coh,
        slc_stack=slc_stack,
        strides=strides,
        avg_mag=avg_mag,
        ps_mask=ps_mask,
        reference_idx=reference_idx,
    )



import numpy as np
import pytest
from scipy.stats import rayleigh

from dolphin import shp
from dolphin.phase_link import simulate
from dolphin.shp._common import remove_unconnected

simulate._seed(1234)


@pytest.mark.parametrize("method", ["glrt", "ks"])
def test_shp_glrt_tf_smoketest(method):
    shape = (5, 50, 50)
    slcs = 20 * (np.random.rand(*shape) + 1j * np.random.rand(*shape))
    amp_stack = np.abs(slcs)
    mean = np.mean(amp_stack, axis=0)
    var = np.var(amp_stack, axis=0)

    # Make sure no errors
    shp.estimate_neighbors(
        mean=mean,
        var=var,
        halfwin_rowcol=(3, 5),
        nslc=slcs.shape[0],
        amp_stack=amp_stack,
        alpha=0.05,
        method=method,
    )


NUM_SLCS = 30


@pytest.fixture(scope="module")
def slcs(shape=(NUM_SLCS, 11, 11)):
    return 20 * (np.random.rand(*shape) + 1j * np.random.rand(*shape))


@pytest.fixture(scope="module")
def mean(slcs):
    amp_stack = np.abs(slcs)
    return np.mean(amp_stack, axis=0)


@pytest.fixture(scope="module")
def var(slcs):
    amp_stack = np.abs(slcs)
    return np.var(amp_stack, axis=0)


def test_shp_glrt(mean, var):
    method = "glrt"

    halfwin_rowcol = (5, 5)  # Looking at the entire stack

    # First try a small alpha
    neighbors = shp.estimate_neighbors(
        mean=mean,
        var=var,
        halfwin_rowcol=halfwin_rowcol,
        nslc=NUM_SLCS,
        alpha=0.005,
        method=method,
    )
    shps_mid_pixel = neighbors[5, 5]
    # Check that at least 120/121 is counted as a neighbor
    assert shps_mid_pixel.sum() >= 120

    # Check the edges are cut off
    top_left = neighbors[0, 0]
    assert top_left[:5, :5].sum() == 5 * 5 - 1
    assert top_left[6:, :].sum() == top_left[:, 6:].sum() == 0


def test_shp_ks(slcs):
    amp_stack = np.abs(slcs)
    halfwin_rowcol = (5, 5)  # Looking at the entire stack

    # First try a tiny alpha
    neighbors = shp.estimate_neighbors(
        amp_stack=amp_stack,
        halfwin_rowcol=halfwin_rowcol,
        nslc=NUM_SLCS,
        alpha=1e-6,
        method="ks",
    )
    shps_mid_pixel = neighbors[5, 5]
    # Check that everything is counted as a neighbor
    assert shps_mid_pixel.sum() == shps_mid_pixel.size - 1

    # Check the edges are cut off and all zeros
    # NOTE: this is different than TF test since KS is slower
    # TODO if i want them to be exactly the same...
    top_left = neighbors[0, 0]
    assert not top_left[:5, :5].any()

    # Next try a large alpha (should be no neighbors)
    neighbors = shp.estimate_neighbors(
        amp_stack=amp_stack,
        halfwin_rowcol=halfwin_rowcol,
        nslc=NUM_SLCS,
        alpha=1.0,
        method="ks",
    )
    shps_mid_pixel = neighbors[5, 5]
    assert shps_mid_pixel.sum() == 0  # only itself


def test_shp_half_mean_different(mean, var):
    """Run a test where half the image has different mean"""
    method = "glrt"

    halfwin_rowcol = (5, 5)
    # make the top half different amplitude
    mean2 = mean.copy()
    mean2[:5, :] += 2000
    # For this test, make all variances equal (just checking mean)
    var[:] = var[5, 5]

    # Even at a small alpha, it should identify the top half as different
    neighbors = shp.estimate_neighbors(
        mean=mean2,
        var=var,
        halfwin_rowcol=halfwin_rowcol,
        nslc=NUM_SLCS,
        alpha=0.01,
        method=method,
    )
    shps_mid_pixel = neighbors[5, 5]
    # Check that everything is counted as a neighbor
    top_block = shps_mid_pixel[5:, :]
    assert top_block.sum() == top_block.size - 1
    assert not shps_mid_pixel[:5, :].any()


def test_shp_half_var_different(mean, var):
    """Run a test where half the image has different variance"""
    method = "glrt"

    halfwin_rowcol = (5, 5)
    # make the top half different amplitude
    var2 = var.copy()
    var2[:5, :] += 5000
    # For this test, make all means equal (just checking var)
    mean[:] = mean[5, 5]

    # Even at a small alpha, it should identify the top half as different
    neighbors = shp.estimate_neighbors(
        mean=mean,
        var=var2,
        halfwin_rowcol=halfwin_rowcol,
        nslc=NUM_SLCS,
        alpha=0.01,
        method=method,
    )
    shps_mid_pixel = neighbors[5, 5]
    top_block = shps_mid_pixel[5:, :]
    assert top_block.sum() == top_block.size - 1
    assert not shps_mid_pixel[:5, :].any()


@pytest.mark.parametrize("strides", [{"x": 1, "y": 1}, {"x": 2, "y": 2}])
def test_shp_glrt_nodata_0(mean, var, strides):
    """Ensure"""
    method = "glrt"

    halfwin_rowcol = (5, 5)  # Looking at the entire stack

    mean = mean.copy()
    mean[:4, :4] = 0
    # First try a small alpha
    neighbors = shp.estimate_neighbors(
        mean=mean,
        var=var,
        halfwin_rowcol=halfwin_rowcol,
        strides=strides,
        nslc=NUM_SLCS,
        alpha=0.005,
        method=method,
    )
    out_col, out_row = 2 // strides["x"], 2 // strides["x"]
    assert neighbors[:out_row, :out_col, :, :].sum() == 0


@pytest.mark.parametrize("method", ["glrt", "ks"])
@pytest.mark.parametrize("alpha", [0.01, 0.05])
@pytest.mark.parametrize("strides", [{"x": 1, "y": 1}, {"x": 2, "y": 2}])
def test_shp_statistics(method, alpha, strides):
    """Check that with repeated tries, the alpha is correct."""

    nsim = 200
    shape = (30, 11, 11)
    halfwin_rowcol = (5, 5)
    shp_counts = np.zeros(nsim)

    amp_stack_sims = rayleigh.rvs(scale=10, size=(nsim, *shape))
    for i in range(nsim):
        amp_stack = amp_stack_sims[i]
        mean = np.mean(amp_stack, axis=0)
        var = np.var(amp_stack, axis=0)
        amp_stack.shape[0]

        neighbors = shp.estimate_neighbors(
            mean=mean,
            var=var,
            halfwin_rowcol=halfwin_rowcol,
            nslc=NUM_SLCS,
            strides=strides,
            alpha=alpha,
            method=method,
            amp_stack=amp_stack,
        )

        out_rows, out_cols = neighbors.shape[:2]
        shps_mid_pixel = neighbors[out_rows // 2, out_cols // 2]
        # dont count center (self) as one
        shp_counts[i] = shps_mid_pixel.sum() - 1

    # Check that the mean number of SHPs is close to 5%
    nbox = shape[1] * shape[2] - 1
    shp_frac = shp_counts.mean() / nbox
    tol_pct = 3
    assert 100 * np.abs(shp_frac - (1 - alpha)) < tol_pct


def test_remove_unconnected():
    # Test 1: All True values
    data = np.ones((5, 5), dtype=bool)
    result = remove_unconnected(data, inplace=False)
    assert np.all(result == data)

    # Resused in all other tests:
    center_only = np.zeros((5, 5), dtype=bool)
    center_only[2, 2] = True

    # Test 2: All False values except center
    data = center_only.copy()
    result = remove_unconnected(data, inplace=False)
    assert np.all(result == center_only)

    # Test 3: Single unconnected True value
    data = np.zeros((5, 5), dtype=bool)
    data[2, 2] = data[0, 0] = True

    result = remove_unconnected(data, inplace=False)
    assert np.all(result == center_only)

    # Test 4: Single unconnected True value with a connected True group
    data = center_only.copy()
    data[3, 3] = data[0, 0] = True

    expected = np.zeros((5, 5), dtype=bool)
    expected[2, 2] = expected[3, 3] = True
    result = remove_unconnected(data, inplace=False)
    assert np.all(result == expected)

    # Test 5: Inplace modification
    data = np.zeros((5, 5), dtype=bool)
    data[2, 2] = data[3, 3] = data[0, 0] = True
    expected = np.zeros((5, 5), dtype=bool)
    expected[2, 2] = expected[3, 3] = True
    remove_unconnected(data, inplace=True)
    assert np.all(data == expected)



from __future__ import annotations

import copy
import logging
import re
from pathlib import Path
from typing import Union
from urllib.parse import ParseResult, urlparse

from dolphin._types import GeneralPath

__all__ = ["S3Path"]


logger = logging.getLogger(__name__)


class S3Path(GeneralPath):
    """A convenience class to handle paths on S3.

    This class relies on `pathlib.Path` for operations using `urllib` to parse the url.

    If passing a url with a trailing slash, the slash will be preserved
    when converting back to string.

    Note that pure path manipulation functions do *not* require `boto3`,
    but functions which interact with S3 (e.g. `exists()`, `.read_text()`) do.

    Attributes
    ----------
    bucket : str
        Name of bucket in the url
    path : pathlib.Path
        The URL path after s3://<bucket>/
    key : str
        Alias of `path` converted to a string

    Examples
    --------
    >>> from dolphin.io import S3Path
    >>> s3_path = S3Path("s3://bucket/path/to/file.txt")
    >>> str(s3_path)
    's3://bucket/path/to/file.txt'
    >>> s3_path.parent
    S3Path("s3://bucket/path/to/")
    >>> str(s3_path.parent)
    's3://bucket/path/to/'

    """

    def __init__(self, s3_url: Union[str, "S3Path"], unsigned: bool = False):
        """Create an S3Path.

        Parameters
        ----------
        s3_url : str or S3Path
            The S3 url to parse.
        unsigned : bool, optional
            If True, disable signing requests to S3.

        """
        # Names come from the urllib.parse.ParseResult
        if isinstance(s3_url, S3Path):
            self._scheme: str = s3_url._scheme
            self._netloc: str = s3_url._netloc
            self.bucket: str = s3_url.bucket
            self.path: Path = s3_url.path
            self._trailing_slash: str = s3_url._trailing_slash
        else:
            parsed: ParseResult = urlparse(s3_url)
            self._scheme = parsed.scheme
            self._netloc = self.bucket = parsed.netloc
            self._parsed = parsed
            self.path = Path(parsed.path)
            self._trailing_slash = "/" if s3_url.endswith("/") else ""

        if self._scheme != "s3":
            raise ValueError(f"{s3_url} is not an S3 url")

        self._unsigned = unsigned

    @classmethod
    def from_bucket_key(cls, bucket: str, key: str):
        """Create a `S3Path` from the bucket name and key/prefix.

        Matches API of some Boto3 functions which use this format.

        Parameters
        ----------
        bucket : str
            Name of S3 bucket.
        key : str
            S3 url of path after the bucket.

        """
        return cls(f"s3://{bucket}/{key}")

    def get_path(self):
        # For S3 paths, we need to add the double slash and netloc back to the front
        return f"{self._scheme}://{self._netloc}{self.path.as_posix()}{self._trailing_slash}"

    @property
    def key(self) -> str:
        """Name of key/prefix within the bucket with leading slash removed."""
        return f"{str(self.path.as_posix()).lstrip('/')}{self._trailing_slash}"

    @property
    def parent(self):
        parent_path = self.path.parent
        # Since this is a parent, it will will always end in a slash
        if self._scheme == "s3":
            # For S3 paths, we need to add the scheme and netloc back to the front
            return S3Path(f"{self._scheme}://{self._netloc}{parent_path.as_posix()}/")
        else:
            # For local paths, we can just convert the path to a string
            return S3Path(str(parent_path) + "/")

    @property
    def suffix(self):
        return self.path.suffix

    def resolve(self) -> S3Path:
        """Resolve the path to an absolute path- S3 paths are always absolute."""
        return self

    def _get_client(self):
        import boto3
        from botocore import UNSIGNED
        from botocore.config import Config

        if self._unsigned:
            return boto3.client("s3", config=Config(signature_version=UNSIGNED))
        else:
            return boto3.client("s3")

    def exists(self) -> bool:
        """Whether this path exists on S3."""
        client = self._get_client()
        resp = client.list_objects_v2(
            Bucket=self.bucket,
            Prefix=self.key,
            MaxKeys=1,
        )
        return resp.get("KeyCount") == 1

    def read_text(self) -> str:
        """Download/read the S3 file as text."""
        return self._download_as_bytes().decode()

    def read_bytes(self) -> bytes:
        """Download/read the S3 file as bytes."""
        return self._download_as_bytes()

    def _download_as_bytes(self) -> bytes:
        """Download file to a `BytesIO` buffer to read as bytes."""
        from io import BytesIO

        client = self._get_client()

        bio = BytesIO()
        client.download_fileobj(self.bucket, self.key, bio)
        bio.seek(0)
        out = bio.read()
        bio.close()
        return out

    def __truediv__(self, other):
        new = copy.deepcopy(self)
        new.path = self.path / other
        new._trailing_slash = "/" if str(other).endswith("/") else ""
        return new

    def __eq__(self, other):
        if isinstance(other, S3Path):
            return self.get_path() == other.get_path()
        elif isinstance(other, str):
            return self.get_path() == other
        else:
            return False

    def __repr__(self):
        return f'S3Path("{self.get_path()}")'

    def __str__(self):
        return self.get_path()

    def to_gdal(self):
        """Convert this S3Path to a GDAL URL."""
        return f"/vsis3/{self.bucket}/{self.key}"


def fix_s3_url(url):
    """Fix an S3 URL that has been altered by pathlib.

    Will replace s3:/my-bucket/... with s3://my-bucket/...
    """
    return re.sub(r"s3:/((?!/).*)", r"s3://\1", str(url))



import json
import logging
import logging.config
import os
import sys
import time
from collections.abc import Callable
from datetime import datetime, timezone
from functools import wraps
from pathlib import Path

from dolphin._types import P, PathOrStr, T

LOG_RECORD_BUILTIN_ATTRS = {
    "args",
    "asctime",
    "created",
    "exc_info",
    "exc_text",
    "filename",
    "funcName",
    "levelname",
    "levelno",
    "lineno",
    "module",
    "msecs",
    "message",
    "msg",
    "name",
    "pathname",
    "process",
    "processName",
    "relativeCreated",
    "stack_info",
    "thread",
    "threadName",
    "taskName",
}
__all__ = ["setup_logging", "log_runtime"]


def setup_logging(
    *,
    logger_name: str = "dolphin",
    debug: bool = False,
    filename: PathOrStr | None = None,
):
    config_file = Path(__file__).parent / Path("log-config.json")
    with open(config_file) as f_in:
        config = json.load(f_in)

    if logger_name not in config["loggers"]:
        config["loggers"][logger_name] = {"level": "INFO", "handlers": ["stderr"]}

    if debug:
        config["loggers"][logger_name]["level"] = "DEBUG"

    if filename:
        if "file" not in config["loggers"][logger_name]["handlers"]:
            config["loggers"][logger_name]["handlers"].append("file")
        config["handlers"]["file"]["filename"] = os.fspath(filename)
        Path(filename).parent.mkdir(exist_ok=True, parents=True)

    if "filename" not in config["handlers"]["file"]:
        # We never passed in a filename: don't log to a file
        config["handlers"].pop("file")

    logging.config.dictConfig(config)

    # Temp work around for tqdm on py312
    if sys.version_info.major == 3 and sys.version_info.minor == 12:
        os.environ["TQDM_DISABLE"] = "1"


def log_runtime(f: Callable[P, T]) -> Callable[P, T]:
    """Decorate a function to time how long it takes to run.

    Usage
    -----
    @log_runtime
    def test_func():
        return 2 + 4
    """
    logger = logging.getLogger(__name__)

    @wraps(f)
    def wrapper(*args: P.args, **kwargs: P.kwargs):
        t1 = time.time()

        result = f(*args, **kwargs)

        t2 = time.time()
        elapsed_seconds = t2 - t1
        elapsed_minutes = elapsed_seconds / 60.0
        time_string = (
            f"Total elapsed time for {f.__module__}.{f.__name__} : "
            f"{elapsed_minutes:.2f} minutes ({elapsed_seconds:.2f} seconds)"
        )

        logger.info(time_string)

        return result

    return wrapper


class JSONFormatter(logging.Formatter):
    def __init__(
        self,
        *,
        fmt_keys: dict[str, str] | None = None,
    ):
        super().__init__()
        self.fmt_keys = fmt_keys if fmt_keys is not None else {}

    def format(self, record: logging.LogRecord) -> str:
        message = self._prepare_log_dict(record)
        return json.dumps(message, default=str)

    def _prepare_log_dict(self, record: logging.LogRecord):
        always_fields = {
            "timestamp": datetime.fromtimestamp(
                record.created, tz=timezone.utc
            ).isoformat(),
            "message": record.getMessage(),
        }
        if record.exc_info is not None:
            always_fields["exc_info"] = self.formatException(record.exc_info)

        if record.stack_info is not None:
            always_fields["stack_info"] = self.formatStack(record.stack_info)

        message = {
            key: (
                msg_val
                if (msg_val := always_fields.pop(val, None)) is not None
                else getattr(record, val)
            )
            for key, val in self.fmt_keys.items()
        }
        message.update(always_fields)

        for key, val in record.__dict__.items():
            if key not in LOG_RECORD_BUILTIN_ATTRS:
                message[key] = val

        return message



import threading
import time
from pathlib import Path

from dolphin._decorators import atomic_output


def _long_write(filename, pause: float = 0.2):
    """Simulate a long writing process"""
    with open(filename, "w") as f:
        f.write("aaa\n")
        time.sleep(pause)
        f.write("bbb\n")


@atomic_output(output_arg="outname")
def default_write_newname(outname="out.txt"):
    _long_write(outname)


@atomic_output(output_arg="outname", use_tmp=True)
def default_write_newname_tmp(outname="out.txt"):
    _long_write(outname)


@atomic_output(output_arg="output_dir", is_dir=True)
def default_write_dir(output_dir="some_dir", filename="testfile.txt"):
    p = Path(output_dir)
    p.mkdir(exist_ok=True, parents=True)
    outname = p / filename
    _long_write(outname)


@atomic_output(output_arg="output_dir", is_dir=True, use_tmp=True)
def default_write_dir_tmp(output_dir="some_dir", filename="testfile.txt"):
    p = Path(output_dir)
    p.mkdir(exist_ok=True, parents=True)
    outname = p / filename
    _long_write(outname)


def test_atomic_output(tmpdir):
    with tmpdir.as_cwd():
        default_write_newname(outname="out1.txt")
        default_write_newname(outname="out2.txt")
        for fn in ["out1.txt", "out2.txt"]:
            assert Path(fn).exists()


def test_atomic_output_tmp(tmpdir):
    with tmpdir.as_cwd():
        default_write_newname_tmp(outname="out1.txt")
        assert Path("out1.txt").exists()


def test_atomic_output_dir(tmp_path: Path):
    out_dir = tmp_path / "out"
    filename = "testfile.txt"
    out_dir.mkdir()
    default_write_dir(output_dir=out_dir, filename=filename)
    assert Path(out_dir / filename).exists()


def test_atomic_output_dir_tmp(tmp_path: Path):
    out_dir = tmp_path / "out"
    filename = "testfile.txt"
    out_dir.mkdir()
    default_write_dir(output_dir=out_dir, filename=filename)
    assert Path(out_dir / filename).exists()


def test_atomic_output_name_swap_file(tmpdir):
    with tmpdir.as_cwd():
        outname2 = "out3.txt"
        t = threading.Thread(target=default_write_newname, kwargs={"outname": outname2})
        t.start()
        # It should NOT exist, yet
        assert not Path(outname2).exists()
        time.sleep(0.5)
        t.join()
        assert Path(outname2).exists()


def test_atomic_output_dir_swap(tmp_path: Path):
    # Kick off the writing function in the background
    # so we see if a different file was created
    # Check it works providing the "args"
    out_dir = tmp_path / "out"
    out_dir.mkdir()
    t = threading.Thread(target=default_write_dir, kwargs={"output_dir": out_dir})
    t.start()
    # It should NOT exist, yet
    assert not Path(out_dir / "testfile.txt").exists()
    time.sleep(0.5)
    t.join()
    assert Path(out_dir / "testfile.txt").exists()



"""Stitch burst interferograms (optional) and unwrap them."""

from __future__ import annotations

import logging
import time
from pathlib import Path
from typing import Sequence

from dolphin import io, stitching, unwrap
from dolphin._log import log_runtime
from dolphin._overviews import ImageType, create_overviews
from dolphin._types import PathOrStr

from .config import UnwrapOptions

logger = logging.getLogger(__name__)


@log_runtime
def run(
    ifg_file_list: Sequence[Path],
    cor_file_list: Sequence[Path],
    nlooks: float,
    unwrap_options: UnwrapOptions,
    temporal_coherence_file: PathOrStr | None = None,
    mask_file: PathOrStr | None = None,
    add_overviews: bool = True,
) -> tuple[list[Path], list[Path]]:
    """Run the displacement workflow on a stack of SLCs.

    Parameters
    ----------
    ifg_file_list : Sequence[Path]
        Sequence interferograms files to unwrap.
    cor_file_list : Sequence[Path]
        Sequence interferometric correlation files, one per file in `ifg_file_list`
    nlooks : float
        Effective number of looks used to form the input correlation data.
    unwrap_options : UnwrapOptions
        [`UnwrapOptions`][dolphin.workflows.config.UnwrapOptions] config object
        with parameters for running unwrapping jobs.
    temporal_coherence_file : Filename, optional
        Path to temporal coherence file from phase linking.
    mask_file : PathOrStr, optional
        Path to boolean mask indicating nodata areas.
        1 indicates valid data, 0 indicates missing data.
    add_overviews : bool, default = True
        If True, creates overviews of the unwrapped phase and connected component
        labels.

    Returns
    -------
    unwrapped_paths : list[Path]
        list of Paths to unwrapped interferograms created.
    conncomp_paths : list[Path]
        list of Paths to connected component files created.

    """
    t0 = time.perf_counter()
    if len(ifg_file_list) != len(cor_file_list):
        msg = f"{len(ifg_file_list) = } != {len(cor_file_list) = }"
        raise ValueError(msg)

    output_path = unwrap_options._directory
    output_path.mkdir(exist_ok=True, parents=True)
    if mask_file is not None:
        output_mask = _get_matching_mask(
            mask_file=mask_file,
            output_dir=output_path,
            match_file=ifg_file_list[0],
        )
    else:
        output_mask = None

    logger.info(f"Unwrapping {len(ifg_file_list)} interferograms")

    # Make a scratch directory for unwrapping
    unwrap_scratchdir = unwrap_options._directory / "scratch"
    unwrap_scratchdir.mkdir(exist_ok=True, parents=True)

    unwrapped_paths, conncomp_paths = unwrap.run(
        ifg_filenames=ifg_file_list,
        cor_filenames=cor_file_list,
        output_path=output_path,
        unwrap_options=unwrap_options,
        nlooks=nlooks,
        temporal_coherence_file=temporal_coherence_file,
        mask_filename=output_mask,
        scratchdir=unwrap_scratchdir,
    )

    if add_overviews:
        logger.info("Creating overviews for unwrapped images")
        create_overviews(unwrapped_paths, image_type=ImageType.UNWRAPPED)
        create_overviews(conncomp_paths, image_type=ImageType.CONNCOMP)

    # Dump the used options for JSON parsing
    logger.info(
        "unwrapping complete",
        extra={
            "elapsed": time.perf_counter() - t0,
            "unwrap_options": unwrap_options.model_dump(mode="json"),
        },
    )

    return (unwrapped_paths, conncomp_paths)


def _get_matching_mask(
    mask_file: PathOrStr, output_dir: Path, match_file: PathOrStr
) -> Path:
    """Create a mask with the same size/projection as `match_file`."""
    # Check that the input mask is the same size as the ifgs:
    if io.get_raster_xysize(mask_file) == match_file:
        logger.info(f"Using {mask_file} to mask during unwrapping")
        output_mask = Path(mask_file)
    else:
        logger.info(f"Warping {mask_file} to match size of interferograms")
        output_mask = output_dir / "warped_mask.tif"
        if output_mask.exists():
            logger.info(f"Mask already exists at {output_mask}")
        else:
            stitching.warp_to_match(
                input_file=mask_file,
                match_file=match_file,
                output_file=output_mask,
            )
    return output_mask



from __future__ import annotations

import datetime
import logging
import math
import resource
import sys
import warnings
from collections.abc import Callable
from concurrent.futures import Executor, Future
from itertools import chain
from multiprocessing import cpu_count
from pathlib import Path
from typing import Any, Iterable, Optional, Sequence, Union

import numpy as np
from numpy.typing import ArrayLike, DTypeLike
from osgeo import gdal, gdal_array, gdalconst

from dolphin._types import Bbox, Filename, P, Strides, T

DateOrDatetime = Union[datetime.date, datetime.datetime]

gdal.UseExceptions()
logger = logging.getLogger(__name__)


def numpy_to_gdal_type(np_dtype: DTypeLike) -> int:
    """Convert numpy dtype to gdal type.

    Parameters
    ----------
    np_dtype : DTypeLike
        Numpy dtype to convert.

    Returns
    -------
    int
        GDAL type code corresponding to `np_dtype`.

    Raises
    ------
    TypeError
        If `np_dtype` is not a numpy dtype, or if the provided dtype is not
        supported by GDAL (for example, `np.dtype('>i4')`)

    """
    np_dtype = np.dtype(np_dtype)

    if np.issubdtype(bool, np_dtype):
        return gdalconst.GDT_Byte
    gdal_code = gdal_array.NumericTypeCodeToGDALTypeCode(np_dtype)
    if gdal_code is None:
        msg = f"dtype {np_dtype} not supported by GDAL."
        raise TypeError(msg)
    return gdal_code


def gdal_to_numpy_type(gdal_type: Union[str, int]) -> np.dtype:
    """Convert gdal type to numpy type."""
    if isinstance(gdal_type, str):
        gdal_type = gdal.GetDataTypeByName(gdal_type)
    return np.dtype(gdal_array.GDALTypeCodeToNumericTypeCode(gdal_type))


def _get_path_from_gdal_str(name: Filename) -> Path:
    s = str(name)
    if s.upper().startswith("DERIVED_SUBDATASET"):
        # like DERIVED_SUBDATASET:AMPLITUDE:slc_filepath.tif
        p = s.split(":")[-1].strip('"').strip("'")
    elif ":" in s and (s.upper().startswith("NETCDF") or s.upper().startswith("HDF")):
        # like NETCDF:"slc_filepath.nc":subdataset
        p = s.split(":")[1].strip('"').strip("'")
    else:
        # Whole thing is the path
        p = str(name)
    return Path(p)


def _resolve_gdal_path(gdal_str: Filename) -> Filename:
    """Resolve the file portion of a gdal-openable string to an absolute path."""
    s_clean = str(gdal_str).lstrip('"').lstrip("'").rstrip('"').rstrip("'")
    prefixes = ["DERIVED_SUBDATASET", "NETCDF", "HDF"]
    is_gdal_str = any(s_clean.upper().startswith(pre) for pre in prefixes)
    file_part = str(_get_path_from_gdal_str(s_clean))

    # strip quotes to add back in after
    file_part = file_part.strip('"').strip("'")
    file_part_resolved = Path(file_part).resolve()
    resolved = s_clean.replace(file_part, str(file_part_resolved))
    return Path(resolved) if not is_gdal_str else resolved


def _get_slices(half_r: int, half_c: int, r: int, c: int, rows: int, cols: int):
    """Get the slices for the given pixel and half window size."""
    # Clamp min indexes to 0
    r_start = max(r - half_r, 0)
    c_start = max(c - half_c, 0)
    # Clamp max indexes to the array size
    r_end = min(r + half_r + 1, rows)
    c_end = min(c + half_c + 1, cols)
    return (r_start, r_end), (c_start, c_end)


def full_suffix(filename: Filename):
    """Get the full suffix of a filename, including multiple dots.

    Parameters
    ----------
    filename : str or Path
        path to file

    Returns
    -------
    str
        The full suffix, including multiple dots.

    Examples
    --------
        >>> full_suffix('test.tif')
        '.tif'
        >>> full_suffix('test.tar.gz')
        '.tar.gz'

    """
    fpath = Path(filename)
    return "".join(fpath.suffixes)


def disable_gpu():
    """Disable GPU usage."""
    import os

    import jax

    os.environ["CUDA_VISIBLE_DEVICES"] = ""
    jax.config.update("jax_platform_name", "cpu")


def gpu_is_available() -> bool:
    """Check if a GPU is available."""
    # TODO: not sure yet how to check for the jax gpu installation
    try:
        from numba import cuda
        from numba.cuda.cudadrv.error import CudaRuntimeError

    except ImportError:
        logger.debug("numba installed, but GPU not available")
        return False
    try:
        cuda_version = cuda.runtime.get_version()
        logger.debug(f"CUDA version: {cuda_version}")
    except (OSError, CudaRuntimeError) as e:
        logger.debug(f"CUDA runtime version error: {e}")
        return False
    try:
        n_gpu = len(cuda.gpus)
    except cuda.CudaSupportError as e:
        logger.debug(f"CUDA support error {e}")
        return False
    if n_gpu < 1:
        logger.debug("No available GPUs found")
        return False
    return True


def take_looks(arr, row_looks, col_looks, func_type="nansum", edge_strategy="cutoff"):
    """Downsample a numpy matrix by summing blocks of (row_looks, col_looks).

    Parameters
    ----------
    arr : np.array
        2D array of an image
    row_looks : int
        the reduction rate in row direction
    col_looks : int
        the reduction rate in col direction
    func_type : str, optional
        the numpy function to use for downsampling, by default "nansum"
    edge_strategy : str, optional
        how to handle edges, by default "cutoff"
        Choices are "cutoff", "pad"

    Returns
    -------
    ndarray
        The downsampled array, shape = ceil(rows / row_looks, cols / col_looks)

    Notes
    -----
    Cuts off values if the size isn't divisible by num looks.

    """
    if row_looks == 1 and col_looks == 1:
        return arr

    if arr.ndim >= 3:
        return np.stack(
            [
                take_looks(
                    a,
                    row_looks,
                    col_looks,
                    func_type=func_type,
                    edge_strategy=edge_strategy,
                )
                for a in arr
            ]
        )

    if isinstance(arr, np.ma.MaskedArray):
        # Must do looks separately on mask
        # https://github.com/numpy/numpy/issues/8881
        looked_data = take_looks(
            arr.data,
            row_looks,
            col_looks,
            func_type=func_type,
            edge_strategy=edge_strategy,
        )
        if arr.mask.ndim == np.ma.nomask:
            looked_mask = arr.mask
        else:
            looked_mask = take_looks(
                arr.mask,
                row_looks,
                col_looks,
                func_type="any",
                edge_strategy=edge_strategy,
            )
        return np.ma.MaskedArray(data=looked_data, mask=looked_mask)

    arr = _make_dims_multiples(arr, row_looks, col_looks, how=edge_strategy)

    rows, cols = arr.shape
    new_rows = rows // row_looks
    new_cols = cols // col_looks

    func = getattr(np, func_type)
    with warnings.catch_warnings():
        # ignore the warning about nansum of empty slice
        warnings.simplefilter("ignore", category=RuntimeWarning)
        return func(
            np.reshape(arr, (new_rows, row_looks, new_cols, col_looks)), axis=(3, 1)
        )


def _make_dims_multiples(arr, row_looks, col_looks, how="cutoff"):
    """Pad or cutoff an array to make the dimensions multiples of the looks."""
    rows, cols = arr.shape
    row_cutoff = rows % row_looks
    col_cutoff = cols % col_looks
    if how == "cutoff":
        if row_cutoff != 0:
            arr = arr[:-row_cutoff, :]
        if col_cutoff != 0:
            arr = arr[:, :-col_cutoff]
        return arr
    elif how == "pad":
        pad_rows = (row_looks - row_cutoff) % row_looks
        pad_cols = (col_looks - col_cutoff) % col_looks
        pad_val = False if arr.dtype == bool else np.nan
        if pad_rows > 0 or pad_cols > 0:
            arr = np.pad(
                arr,
                ((0, pad_rows), (0, pad_cols)),
                mode="constant",
                constant_values=pad_val,
            )
        return arr
    else:
        msg = f"Invalid edge strategy: {how}"
        raise ValueError(msg)


def upsample_nearest(
    arr: np.ndarray,
    output_shape: tuple[int, int],
    looks: Optional[tuple[int, int]] = None,
) -> np.ndarray:
    """Upsample a numpy matrix by repeating blocks of (row_looks, col_looks).

    Parameters
    ----------
    arr : np.array
        2D or 3D downsampled array.
    output_shape : tuple[int, int]
        The desired output shape.
    looks : tuple[int, int]
        The number of looks in the row and column directions.
        If not provided, will be calculated from `output_shape`.

    Returns
    -------
    ndarray
        The upsampled array, shape = `output_shape`.

    """
    in_rows, in_cols = arr.shape[-2:]
    out_rows, out_cols = output_shape[-2:]
    if (in_rows, in_cols) == (out_rows, out_cols):
        return arr

    if looks is None:
        row_looks = out_rows // in_rows
        col_looks = out_cols // in_cols
    else:
        row_looks, col_looks = looks

    arr_up = np.repeat(np.repeat(arr, row_looks, axis=-2), col_looks, axis=-1)
    # This may be larger than the original array, or it may be smaller, depending
    # on whether it was padded or cutoff
    out_r = min(out_rows, arr_up.shape[-2])
    out_c = min(out_cols, arr_up.shape[-1])

    shape = (len(arr), out_rows, out_cols) if arr.ndim == 3 else (out_rows, out_cols)
    arr_out = np.zeros(shape=shape, dtype=arr.dtype)
    arr_out[..., :out_r, :out_c] = arr_up[..., :out_r, :out_c]
    return arr_out


def get_max_memory_usage(units: str = "GB", children: bool = True) -> float:
    """Get the maximum memory usage of the current process.

    Parameters
    ----------
    units : str, optional, choices=["GB", "MB", "KB", "byte"]
        The units to return, by default "GB".
    children : bool, optional
        Whether to include the memory usage of child processes, by default True

    Returns
    -------
    float
        The maximum memory usage in the specified units.

    Raises
    ------
    ValueError
        If the units are not recognized.

    References
    ----------
    1. https://stackoverflow.com/a/7669279/4174466
    2. https://unix.stackexchange.com/a/30941/295194
    3. https://manpages.debian.org/bullseye/manpages-dev/getrusage.2.en.html

    """
    max_mem = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
    if children:
        max_mem += resource.getrusage(resource.RUSAGE_CHILDREN).ru_maxrss
    if units.lower().startswith("g"):
        factor = 1e9
    elif units.lower().startswith("m"):
        factor = 1e6
    elif units.lower().startswith("k"):
        factor = 1e3
    elif units.lower().startswith("byte"):
        factor = 1.0
    else:
        msg = f"Unknown units: {units}"
        raise ValueError(msg)
    if sys.platform.startswith("linux"):
        # on linux, ru_maxrss is in kilobytes, while on mac, ru_maxrss is in bytes
        factor /= 1e3

    return max_mem / factor


def get_gpu_memory(pid: Optional[int] = None, gpu_id: int = 0) -> float:
    """Get the memory usage (in GiB) of the GPU for the current pid."""
    try:
        from pynvml.smi import nvidia_smi
    except ImportError as e:
        msg = "Please install pynvml through pip or conda"
        raise ImportError(msg) from e

    def get_mem(process):
        used_mem = process["used_memory"] if process else 0
        if process["unit"] == "MiB":
            multiplier = 1 / 1024
        else:
            logger.warning(f"Unknown unit: {process['unit']}")
        return used_mem * multiplier

    nvsmi = nvidia_smi.getInstance()
    processes = nvsmi.DeviceQuery()["gpu"][gpu_id]["processes"]
    if not processes:
        return 0.0

    if pid is None:
        # Return sum of all processes
        return sum(get_mem(p) for p in processes)
    else:
        procs = [p for p in processes if p["pid"] == pid]
        return get_mem(procs[0]) if procs else 0.0


def moving_window_mean(
    image: ArrayLike, size: Union[int, tuple[int, int]]
) -> np.ndarray:
    """DEPRECATED: use `scipy.ndimage.uniform_filter` directly."""  # noqa: D401
    from scipy.ndimage import uniform_filter

    msg = (
        "`moving_window_mean` is deprecated. Please use `scipy.ndimage.uniform_filter`."
    )
    warnings.warn(msg, category=DeprecationWarning, stacklevel=2)
    return uniform_filter(image, size=size)


def set_num_threads(num_threads: int):
    """Set the cap on threads spawned by numpy and numba.

    Uses https://github.com/joblib/threadpoolctl for numpy.
    """
    import os

    import numba
    from threadpoolctl import ThreadpoolController

    # Set the environment variables for the workers
    controller = ThreadpoolController()
    controller.limit(limits=num_threads)
    # https://numba.readthedocs.io/en/stable/user/threading-layer.html#example-of-limiting-the-number-of-threads
    num_cpus = get_cpu_count()
    numba.set_num_threads(min(num_cpus, num_threads))
    # jax setup is harder, for now
    os.environ["XLA_FLAGS"] = f"--xla_force_host_platform_device_count={num_threads}"


def get_cpu_count():
    """Get the number of CPUs available to the current process.

    This function accounts for the possibility of a Docker container with
    limited CPU resources on a larger machine (which is ignored by
    `multiprocessing.cpu_count()`).

    Returns
    -------
    int
        The number of CPUs available to the current process.

    References
    ----------
    1. https://github.com/joblib/loky/issues/111
    2. https://github.com/conan-io/conan/blob/982a97041e1ece715d157523e27a14318408b925/conans/client/tools/oss.py#L27 # noqa

    """  # noqa: E501

    def get_cpu_quota():
        return int(Path("/sys/fs/cgroup/cpu/cpu.cfs_quota_us").read_text())

    def get_cpu_period():
        return int(Path("/sys/fs/cgroup/cpu/cpu.cfs_period_us").read_text())

    try:
        cfs_quota_us = get_cpu_quota()
        cfs_period_us = get_cpu_period()
        if cfs_quota_us > 0 and cfs_period_us > 0:
            return int(math.ceil(cfs_quota_us / cfs_period_us))
    except Exception:
        pass
    return cpu_count()


def flatten(list_of_lists: Iterable[Iterable[Any]]) -> chain[Any]:
    """Flatten one level of a nested iterable."""
    return chain.from_iterable(list_of_lists)


def format_date_pair(
    start: DateOrDatetime, end: DateOrDatetime, fmt: str = "%Y%m%d"
) -> str:
    """Format a date pair into a string.

    Parameters
    ----------
    start : DateOrDatetime
        First date or datetime
    end : DateOrDatetime
        Second date or datetime
    fmt : str, optional
        `datetime` formatter pattern.
        Default = "%Y%m%d"

    Returns
    -------
    str
        Formatted date pair.

    """
    return format_dates(start, end, fmt=fmt, sep="_")


def format_dates(*dates: DateOrDatetime, fmt: str = "%Y%m%d", sep: str = "_") -> str:
    """Format a date pair into a string.

    Parameters
    ----------
    *dates : DateOrDatetime
        Sequence of date/datetimes to format
    fmt : str, optional
        `datetime` formatter pattern.
        Default = "%Y%m%d"
    sep : str, optional
        string separator between dates.
        Default = "_"

    Returns
    -------
    str
        Formatted date pair.

    """
    return sep.join((d.strftime(fmt)) for d in dates)


# Keep alias for now, but deprecate
_format_date_pair = format_date_pair


def prepare_geometry(
    geometry_dir: Path,
    geo_files: Sequence[Path],
    matching_file: Path,
    dem_file: Optional[Path],
    epsg: int,
    out_bounds: Bbox,
    strides: Optional[dict[str, int]] = None,
) -> dict[str, Path]:
    """Prepare geometry files.

    Parameters
    ----------
    geometry_dir : Path
        Output directory for geometry files.
    geo_files : list[Path]
        list of geometry files.
    matching_file : Path
        Matching file.
    dem_file : Optional[Path]
        DEM file.
    epsg : int
        EPSG code.
    out_bounds : Bbox
        Output bounds.
    strides : Dict[str, int], optional
        Strides for resampling, by default {"x": 1, "y": 1}.

    Returns
    -------
    Dict[str, Path]
        Dictionary of prepared geometry files.

    """
    from dolphin import stitching
    from dolphin.io import format_nc_filename

    if strides is None:
        strides = {"x": 1, "y": 1}
    geometry_dir.mkdir(exist_ok=True)

    stitched_geo_list = {}

    if geo_files[0].name.endswith(".h5"):
        # ISCE3 geocoded SLCs
        datasets = ["los_east", "los_north"]

        for ds_name in datasets:
            outfile = geometry_dir / f"{ds_name}.tif"
            logger.info(f"Creating {outfile}")
            stitched_geo_list[ds_name] = outfile
            ds_path = f"/data/{ds_name}"
            cur_files = [format_nc_filename(f, ds_name=ds_path) for f in geo_files]

            no_data = 0

            stitching.merge_images(
                cur_files,
                outfile=outfile,
                driver="GTiff",
                out_bounds=out_bounds,
                out_bounds_epsg=epsg,
                in_nodata=no_data,
                out_nodata=no_data,
                target_aligned_pixels=True,
                strides=strides,
                overwrite=False,
            )

        if dem_file:
            height_file = geometry_dir / "height.tif"
            stitched_geo_list["height"] = height_file
            if not height_file.exists():
                logger.info(f"Creating {height_file}")
                stitching.warp_to_match(
                    input_file=dem_file,
                    match_file=matching_file,
                    output_file=height_file,
                    resample_alg="cubic",
                )
    else:
        # ISCE2 radar coordinates
        dsets = {
            "hgt.rdr": "height",
            "incLocal.rdr": "incidence_angle",
            "lat.rdr": "latitude",
            "lon.rdr": "longitude",
        }

        for geo_file in geo_files:
            if geo_file.stem in dsets:
                out_name = dsets[geo_file.stem]
            elif geo_file.name in dsets:
                out_name = dsets[geo_file.name]
                continue

            out_file = geometry_dir / (out_name + ".tif")
            stitched_geo_list[out_name] = out_file
            logger.info(f"Creating {out_file}")

            stitching.warp_to_match(
                input_file=geo_file,
                match_file=matching_file,
                output_file=out_file,
                resample_alg="cubic",
            )

    return stitched_geo_list


def compute_out_shape(shape: tuple[int, int], strides: Strides) -> tuple[int, int]:
    """Calculate the output size for an input `shape` and row/col `strides`.

    Parameters
    ----------
    shape : tuple[int, int]
        Input size: (rows, cols)
    strides : tuple[int, int]
        (y strides, x strides)

    Returns
    -------
    out_shape : tuple[int, int]
        Size of output after striding

    Notes
    -----
    If there is not a full window (of size `strides`), the end
    will get cut off rather than padded with a partial one.
    This should match the output size of `[dolphin.utils.take_looks][]`.

    As a 1D example, in array of size 6 with `strides`=3 along this dim,
    we could expect the pixels to be centered on indexes
    `[1, 4]`.

        [ 0  1  2   3  4  5]

    So the output size would be 2, since we have 2 full windows.
    If the array size was 7 or 8, we would have 2 full windows and 1 partial,
    so the output size would still be 2.

    """
    rows, cols = shape
    rstride, cstride = strides
    return (rows // rstride, cols // cstride)


class DummyProcessPoolExecutor(Executor):
    """Dummy ProcessPoolExecutor for to avoid forking for single_job purposes."""

    def __init__(self, max_workers: Optional[int] = None, **kwargs):  # noqa: D107
        self._max_workers = max_workers

    def submit(  # noqa: D102
        self, fn: Callable[P, T], /, *args: P.args, **kwargs: P.kwargs
    ) -> Future[T]:
        future: Future = Future()
        result = fn(*args, **kwargs)
        future.set_result(result)
        return future

    def shutdown(self, wait: bool = True, cancel_futures: bool = True):  # noqa:D102
        pass

    def map(self, fn: Callable[P, T], *iterables, **kwargs):  # noqa: D102
        return map(fn, *iterables)



import argparse
from pathlib import Path
from typing import TYPE_CHECKING, Any

from dolphin.workflows.config import UnwrapMethod

if TYPE_CHECKING:
    _SubparserType = argparse._SubParsersAction[argparse.ArgumentParser]
else:
    _SubparserType = Any


def get_parser(subparser=None, subcommand_name="unwrap") -> argparse.ArgumentParser:
    """Set up the command line interface."""
    metadata = {
        "description": "Create a configuration file for a displacement workflow.",
        "formatter_class": argparse.ArgumentDefaultsHelpFormatter,
        # https://docs.python.org/3/library/argparse.html#fromfile-prefix-chars
        "fromfile_prefix_chars": "@",
    }
    if subparser:
        # Used by the subparser to make a nested command line interface
        parser = subparser.add_parser(subcommand_name, **metadata)
    else:
        parser = argparse.ArgumentParser(**metadata)  # type: ignore[arg-type]

    # parser._action_groups.pop()
    parser.add_argument(
        "-o",
        "--output-path",
        default=Path(),
        help="Path to output directory to store results",
    )
    # Get Inputs from the command line
    inputs = parser.add_argument_group("Input options")
    inputs.add_argument(
        "--ifg-filenames",
        nargs=argparse.ZERO_OR_MORE,
        help=(
            "List the paths of all ifg files to include. Can pass a newline delimited"
            " file with @ifg_filelist.txt"
        ),
    )
    inputs.add_argument(
        "--cor-filenames",
        nargs=argparse.ZERO_OR_MORE,
        help=(
            "List the paths of all ifg files to include. Can pass a newline delimited"
            " file with @cor_filelist.txt"
        ),
    )
    inputs.add_argument(
        "--mask-filename",
        help=(
            "Path to Byte mask file used to ignore low correlation/bad data (e.g water"
            " mask). Convention is 0 for no data/invalid, and 1 for good data."
        ),
    )
    inputs.add_argument(
        "--temp-coh-filename",
        help="Path to temporal coherence file from phase linking",
    )
    parser.add_argument(
        "--nlooks",
        type=int,
        help="Effective number of looks used to form correlation",
    )

    parser.add_argument(
        "--max-jobs",
        type=int,
        default=1,
        help="Number of parallel files to unwrap",
    )

    algorithm_opts = parser.add_argument_group("Algorithm options")
    algorithm_opts.add_argument(
        "--unwrap-method",
        type=UnwrapMethod,
        choices=[m.value for m in UnwrapMethod],
        default=UnwrapMethod.SNAPHU.value,
        help="Choice of unwrapping algorithm to use.",
    )
    algorithm_opts.add_argument(
        "--run-goldstein",
        action="store_true",
        help="Run Goldstein filter before unwrapping.",
    )
    algorithm_opts.add_argument(
        "--run-interpolation",
        action="store_true",
        help="Run interpolation before unwrapping.",
    )

    spurt_opts = parser.add_argument_group("Spurt options")
    spurt_opts.add_argument(
        "--temp-coh-threshold",
        type=float,
        help="Cutoff on temporal_coherence raster to choose pixels for unwrapping.",
    )

    tophu_opts = parser.add_argument_group("Tophu options")
    # Add ability for downsampling/tiling with tophu
    tophu_opts.add_argument(
        "--ntiles",
        type=int,
        nargs=2,
        metavar=("ROW_TILES", "COL_TILES"),
        default=(1, 1),
        help=(
            "(using tophu) Split the interferograms into this number of tiles along the"
            " (row, col) axis."
        ),
    )
    tophu_opts.add_argument(
        "--downsample-factor",
        type=int,
        nargs=2,
        default=(1, 1),
        help=(
            "(using tophu) Downsample the interferograms by this factor "
            " during multiresolution unwrapping."
        ),
    )
    parser.set_defaults(run_func=_run_unwrap)

    return parser


def _run_unwrap(*args, **kwargs):
    """Run `dolphin.unwrap.run`.

    Wrapper for the dolphin.unwrap to delay import time.
    """
    from dolphin import unwrap

    return unwrap.run(*args, **kwargs)


def main(args=None):
    """Get the command line arguments and unwrap files."""
    from dolphin import unwrap

    parser = get_parser()
    parsed_args = parser.parse_args(args)
    unwrap.run(**vars(parsed_args))


if __name__ == "__main__":
    main()



import os

import numpy as np
import numpy.testing as npt
import pytest

from dolphin._types import HalfWindow, Strides
from dolphin.phase_link import _core, covariance, simulate
from dolphin.utils import gpu_is_available

GPU_AVAILABLE = gpu_is_available() and os.environ.get("NUMBA_DISABLE_JIT") != "1"
NUM_ACQ = 30
simulate._seed(1234)


@pytest.fixture(scope="module")
def slc_samples(C_truth):
    C, _ = C_truth
    ns = 11 * 11
    return simulate.simulate_neighborhood_stack(C, ns)


@pytest.mark.parametrize("baseline_lag", [None, 5])
@pytest.mark.parametrize("use_evd", [False, True])
def test_estimation(C_truth, slc_samples, use_evd, baseline_lag):
    _, truth = C_truth

    C_hat = np.array(covariance.coh_mat_single(slc_samples))
    # Make the single-pixel comparisons with simple implementation
    est_mle_verify = np.angle(simulate.mle(C_hat))
    est_evd_verify = np.angle(simulate.evd(C_hat))

    # Check that the estimates are close to the truth
    err_deg = 10
    assert np.degrees(simulate.rmse(truth, est_evd_verify)) < err_deg
    assert np.degrees(simulate.rmse(truth, est_mle_verify)) < err_deg

    slc_stack = slc_samples.reshape(NUM_ACQ, 11, 11)

    # cpx_phase, temp_coh, eigs, _ = _core.run_cpl(
    pl_out = _core.run_cpl(
        slc_stack,
        HalfWindow(x=5, y=5),
        Strides(x=1, y=1),
        use_evd=use_evd,
        baseline_lag=baseline_lag,
    )
    assert pl_out.cpx_phase.shape == (len(est_mle_verify), 11, 11)
    assert pl_out.temp_coh.shape == (11, 11)
    assert pl_out.eigenvalues.shape == (11, 11)
    if use_evd:
        expected_min_eig = baseline_lag if baseline_lag else NUM_ACQ / 3
        assert np.all(pl_out.eigenvalues > expected_min_eig)
        assert pl_out
    else:
        # should be 1, but floating point rounding sometimes drops
        assert np.all(pl_out.eigenvalues > 0.99)

    # The middle pixel should be the same, since it had the full window
    est_phase = np.angle(pl_out.cpx_phase[:, 5, 5])
    npt.assert_array_almost_equal(est_mle_verify, est_phase, decimal=1)


def test_masked(slc_samples, C_truth):
    slc_stack = slc_samples.copy().reshape(NUM_ACQ, 11, 11)
    mask = np.zeros((11, 11), dtype=bool)
    # Mask the top row
    mask[0, :] = True
    # slc_samples_masked = slc_stack[:, 1:, :].reshape(NUM_ACQ, -1)
    slc_samples_masked = slc_stack[:, ~mask]

    _, truth = C_truth
    C_hat1 = covariance.coh_mat_single(slc_samples_masked)
    est_mle = np.angle(simulate.mle(np.array(C_hat1)))

    err_deg = 10
    assert np.degrees(simulate.rmse(truth, est_mle)) < err_deg

    slc_stack_masked = slc_stack.copy()
    slc_stack_masked[:, mask] = np.nan

    C_hat2 = covariance.estimate_stack_covariance(
        slc_stack_masked, half_window=HalfWindow(5, 5)
    )
    est_full = _core.process_coherence_matrices(C_hat2)[0]
    # Middle pixel should be the same
    npt.assert_array_almost_equal(est_mle, np.angle(est_full[5, 5, :]), decimal=1)


def test_run_phase_linking(slc_samples):
    slc_stack = slc_samples.copy().reshape(NUM_ACQ, 11, 11)
    pl_out = _core.run_phase_linking(
        slc_stack,
        half_window=HalfWindow(5, 5),
    )

    C_hat = covariance.coh_mat_single(slc_samples)
    expected_phase = np.angle(simulate.mle(np.array(C_hat)))

    # Middle pixel should be the same
    npt.assert_array_almost_equal(
        expected_phase, np.angle(pl_out.cpx_phase[:, 5, 5]), decimal=1
    )


@pytest.mark.parametrize("use_slc_amp", [False, True])
@pytest.mark.parametrize("use_max_ps", [False, True])
def test_run_phase_linking_use_slc_amp(slc_samples, use_slc_amp, use_max_ps):
    slc_stack = slc_samples.copy().reshape(NUM_ACQ, 11, 11)
    ps_mask = np.zeros((11, 11), dtype=bool)
    # Specify at least 1 ps
    ps_mask[1, 1] = True
    pl_out = _core.run_phase_linking(
        slc_stack,
        half_window=HalfWindow(5, 5),
        ps_mask=ps_mask,
        use_slc_amp=use_slc_amp,
        use_max_ps=use_max_ps,
    )

    expected = np.abs(slc_stack) if use_slc_amp else 1
    assert np.allclose(np.abs(pl_out.cpx_phase), expected)


def test_run_phase_linking_with_shift(slc_samples):
    slc_stack = slc_samples.copy().reshape(NUM_ACQ, 11, 11)
    # Pretend there's a shift which should lead to nodata at the intersection
    # First row of the first image
    slc_stack[0, 0, :] = np.nan
    # last row of the second image
    slc_stack[1, -1, :] = np.nan
    pl_out = _core.run_phase_linking(
        slc_stack,
        half_window=HalfWindow(5, 5),
    )

    assert np.isnan(pl_out.cpx_phase[:, [0, -1], :]).all()
    assert np.isnan(pl_out.temp_coh[[0, -1], :]).all()

    assert np.all(~np.isnan(pl_out.cpx_phase[:, 1:-1, :]))
    assert np.all(~np.isnan(pl_out.temp_coh[1:-1, :]))


@pytest.mark.parametrize("strides", [1, 2, 3, 4, 5])
@pytest.mark.parametrize("half_window", [5, 11])
@pytest.mark.parametrize("shape", [(10, 11), (15, 20), (50, 75)])
def test_strides_window_sizes(strides, half_window, shape):
    """Test that window and stride sizes don't cause errors for many input shapes."""
    data = np.random.normal(0, 1, size=(4, *shape)).astype(np.complex64)
    if 2 * half_window + 1 > shape[0]:
        # TODO: We should raise an appropriate error, then catch here
        pytest.skip("Window size is too large for input shape")
    _core.run_phase_linking(
        data,
        half_window=HalfWindow(half_window, half_window),
        strides=Strides(strides, strides),
    )


@pytest.mark.parametrize("use_max_ps", [True, False])
@pytest.mark.parametrize("strides", [1, 2, 3])
def test_run_phase_linking_ps_fill(slc_samples, use_max_ps, strides):
    slc_stack = slc_samples.copy().reshape(NUM_ACQ, 11, 11)
    ps_idx = 2
    ps_mask = np.zeros((11, 11), dtype=bool)
    ps_mask[ps_idx, ps_idx] = True
    # Ignore RuntimeWarning
    pl_out = _core.run_phase_linking(
        slc_stack,
        half_window=HalfWindow(5, 5),
        strides=Strides(strides, strides),
        ps_mask=ps_mask,
        use_max_ps=use_max_ps,
    )
    ps_phase = slc_stack[:, ps_idx, ps_idx]
    ps_phase *= ps_phase[0].conj()  # Reference to first acquisition

    out_idx = ps_idx // strides
    npt.assert_array_almost_equal(
        np.angle(ps_phase), np.angle(pl_out.cpx_phase[:, out_idx, out_idx])
    )

    assert pl_out.temp_coh[out_idx, out_idx] == 1



import pytest

from dolphin import stack

# from dolphin._types import HalfWindow, Strides
from dolphin.io import _readers
from dolphin.phase_link import simulate
from dolphin.utils import compute_out_shape, gpu_is_available
from dolphin.workflows import sequential

GPU_AVAILABLE = gpu_is_available()
simulate._seed(1234)


def test_sequential_gtiff(tmp_path, slc_file_list):
    """Run through the sequential estimation with a GeoTIFF stack."""
    vrt_file = tmp_path / "slc_stack.vrt"
    vrt_stack = _readers.VRTStack(slc_file_list, outfile=vrt_file)
    _, rows, cols = vrt_stack.shape

    half_window = {"x": cols // 2, "y": rows // 2}
    strides = {"x": 1, "y": 1}
    out_shape = compute_out_shape((rows, cols), strides=(strides["y"], strides["x"]))
    if not all(out_shape):
        pytest.skip(f"Output shape = {out_shape}")
    output_folder = tmp_path / "sequential"
    ms_size = 10
    ms_planner = stack.MiniStackPlanner(
        file_list=slc_file_list,
        dates=vrt_stack.dates,
        is_compressed=[False] * len(slc_file_list),
        output_folder=output_folder,
    )

    sequential.run_wrapped_phase_sequential(
        slc_vrt_file=vrt_file,
        ministack_planner=ms_planner,
        ministack_size=ms_size,
        half_window=half_window,
        strides=strides,
        ps_mask_file=None,
        amp_mean_file=None,
        amp_dispersion_file=None,
        shp_method="rect",
        shp_alpha=None,
        shp_nslc=None,
    )

    assert len(list(output_folder.glob("2*.slc.tif"))) == vrt_stack.shape[0]


# Input is only (5, 10) so we can't use a larger window.
@pytest.mark.parametrize(
    "half_window, strides",
    [
        ({"x": 1, "y": 1}, {"x": 1, "y": 1}),
        ({"x": 2, "y": 1}, {"x": 3, "y": 2}),
        ({"x": 3, "y": 1}, {"x": 2, "y": 1}),
        # (HalfWindow(1, 1), Strides(1, 1)),
        # (HalfWindow(1, 2), Strides(2, 3)),
        # (HalfWindow(1, 2), Strides(1, 3)),
    ],
)
def test_sequential_nc(tmp_path, slc_file_list_nc, half_window, strides):
    """Check various strides/windows/ministacks with a NetCDF input stack."""
    vrt_file = tmp_path / "slc_stack.vrt"
    v = _readers.VRTStack(slc_file_list_nc, outfile=vrt_file, subdataset="data")

    _, rows, cols = v.shape
    out_shape = compute_out_shape((rows, cols), strides=(strides["y"], strides["x"]))
    if not all(out_shape):
        pytest.skip(f"Output shape = {out_shape}")

    ms_planner = stack.MiniStackPlanner(
        file_list=slc_file_list_nc,
        dates=v.dates,
        is_compressed=[False] * len(slc_file_list_nc),
        output_folder=tmp_path / "sequential",
    )

    sequential.run_wrapped_phase_sequential(
        slc_vrt_file=vrt_file,
        ministack_planner=ms_planner,
        ministack_size=10,
        half_window=half_window,
        strides=strides,
        ps_mask_file=None,
        amp_mean_file=None,
        amp_dispersion_file=None,
        shp_method="rect",
        shp_alpha=None,
        shp_nslc=None,
    )


@pytest.mark.parametrize("ministack_size", [5, 9, 20])
def test_sequential_ministack_sizes(tmp_path, slc_file_list_nc, ministack_size):
    """Check various strides/windows/ministacks with a NetCDF input stack."""
    vrt_file = tmp_path / "slc_stack.vrt"
    # Make it not a round number to test
    vrt_stack = _readers.VRTStack(
        slc_file_list_nc[:21], outfile=vrt_file, subdataset="data"
    )
    _, rows, cols = vrt_stack.shape
    ms_planner = stack.MiniStackPlanner(
        file_list=vrt_stack.file_list,
        dates=vrt_stack.dates,
        is_compressed=[False] * len(vrt_stack.file_list),
        output_folder=tmp_path / "sequential",
    )

    # Record the warning, check after if it's thrown
    sequential.run_wrapped_phase_sequential(
        slc_vrt_file=vrt_file,
        ministack_planner=ms_planner,
        ministack_size=ministack_size,
        half_window={"x": cols // 2, "y": rows // 2},
        strides={"x": 1, "y": 1},
        ps_mask_file=None,
        amp_mean_file=None,
        amp_dispersion_file=None,
        shp_method="rect",
        shp_alpha=None,
        shp_nslc=None,
    )



from __future__ import annotations

import logging
from pathlib import Path
from pprint import pformat

import opera_utils

import dolphin.ps
from dolphin import __version__
from dolphin._log import log_runtime, setup_logging
from dolphin.io import VRTStack
from dolphin.utils import get_max_memory_usage

from .config import PsWorkflow

logger = logging.getLogger(__name__)


@log_runtime
def run(
    cfg: PsWorkflow,
    compute_looked: bool = False,
    debug: bool = False,
) -> list[Path]:
    """Run the displacement workflow on a stack of SLCs.

    Parameters
    ----------
    cfg : PsWorkflow
        [`PsWorkflow`][dolphin.workflows.config.PsWorkflow] object
        for controlling the workflow.
    compute_looked : bool, optional
        Whether to compute the looked version of the PS mask, by default False.
    debug : bool, optional
        Enable debug logging, by default False.

    """
    # Set the logging level for all `dolphin.` modules
    setup_logging(debug=debug, filename=cfg.log_file)
    logger.debug(pformat(cfg.model_dump()))

    output_file_list = [
        cfg.ps_options._output_file,
        cfg.ps_options._amp_mean_file,
        cfg.ps_options._amp_dispersion_file,
    ]
    ps_output = cfg.ps_options._output_file
    if all(f.exists() for f in output_file_list):
        logger.info(f"Skipping making existing PS files {output_file_list}")
        return output_file_list

    # Check the number of bursts that were passed
    try:
        grouped_slc_files = opera_utils.group_by_burst(cfg.cslc_file_list)
        if len(grouped_slc_files) > 1:
            msg = "Multiple bursts not yet supported for PsWorkflow"
            raise NotImplementedError(msg)
    except ValueError as e:
        # Make sure it's not some other ValueError
        if "Could not parse burst id" not in str(e):
            raise
        # Otherwise, we have SLC files which are not OPERA burst files

    # grab the only key (either a burst, or "") and use that
    cfg.create_dir_tree()

    input_file_list = cfg.cslc_file_list
    if not input_file_list:
        msg = "No input files found"
        raise ValueError(msg)

    # #############################################
    # Make a VRT pointing to the input SLC files
    # #############################################
    subdataset = cfg.input_options.subdataset
    vrt_stack = VRTStack(
        input_file_list,
        subdataset=subdataset,
        outfile=cfg.work_directory / "slc_stack.vrt",
    )

    # Make the nodata mask from the polygons, if we're using OPERA CSLCs
    try:
        nodata_mask_file = cfg.work_directory / "nodata_mask.tif"
        opera_utils.make_nodata_mask(
            vrt_stack.file_list, out_file=nodata_mask_file, buffer_pixels=200
        )
    except Exception as e:
        logger.warning(f"Could not make nodata mask: {e}")
        nodata_mask_file = None

    logger.info(f"Creating persistent scatterer file {ps_output}")
    dolphin.ps.create_ps(
        reader=vrt_stack,
        output_file=output_file_list[0],
        output_amp_mean_file=output_file_list[1],
        output_amp_dispersion_file=output_file_list[2],
        like_filename=vrt_stack.outfile,
        amp_dispersion_threshold=cfg.ps_options.amp_dispersion_threshold,
        block_shape=cfg.worker_settings.block_shape,
    )
    # Save a looked version of the PS mask too
    strides = cfg.output_options.strides
    if compute_looked:
        ps_looked_file, amp_disp_looked = dolphin.ps.multilook_ps_files(
            strides=strides,
            ps_mask_file=cfg.ps_options._output_file,
            amp_dispersion_file=cfg.ps_options._amp_dispersion_file,
        )
        output_file_list.append(ps_looked_file)
        output_file_list.append(amp_disp_looked)

    # Print the maximum memory usage for each worker
    max_mem = get_max_memory_usage(units="GB")
    logger.info(f"Maximum memory usage: {max_mem:.2f} GB")
    logger.info(f"Config file dolphin version: {cfg._dolphin_version}")
    logger.info(f"Current running dolphin version: {__version__}")

    return output_file_list



"""Eigenvalue solvers specialized for the phase linking use case.

|                 | EVD               |         | EMI               |         |
| --------------- | ----------------- | ------- | ----------------- | ------- |
|                 | Runtime (seconds) | Speedup | Runtime (seconds) | Speedup |
| `scipy.eigh`    | 185               | -       | 398               | -       |
| Power iteration | 19                | 9.7     | 57                | 6.8     |

"""

from __future__ import annotations

from functools import partial

import jax
import jax.numpy as jnp
from jax import Array, jit, vmap
from jax.lax import while_loop
from jax.typing import ArrayLike

# For both largest and smallest eig, we map over the first two dimensions
# so now instead of one scalar eigenvalue, we have (rows, cols) eigenvalues


@partial(jit, static_argnames=("mu"))
def eigh_smallest_stack(
    C_arrays: ArrayLike,
    mu: float,
) -> tuple[Array, Array]:
    """Get the smallest (eigenvalue, eigenvector) for each pixel in a 3D stack.

    Uses shift inverse iteration to find the eigenvalue closest to `mu`.
    Pick `mu` to be slightly below the smallest eigenvalue for fastest convergence.

    Returns real eigenvalues, assuming the arrays are Hermitian.

    Parameters
    ----------
    C_arrays : ArrayLike
        The stack of coherence matrices.
        Shape = (rows, cols, nslc, nslc)
    mu : float
        The value to use for the shift inverse iteration.
        The eigenvalue closest to this value is returned.
    v0 : ArrayLike, optional
        The initial guess for the eigenvector.
        If None, a vector of 1s is used.
        Shape = (rows, cols, nslc)

    Returns
    -------
    eigenvalues : Array
        The smallest eigenvalue for each pixel's matrix
        Shape = (rows, cols)
    eigenvectors : Array
        The normalized eigenvector corresponding to the smallest eigenvalue
        Shape = (rows, cols, nslc)

    """
    in_axes = (0, None)
    eig_vals, eig_vecs = vmap(
        vmap(inverse_iteration, in_axes=in_axes), in_axes=in_axes
    )(C_arrays, mu)
    return eig_vals.real, eig_vecs


@jit
def eigh_largest_stack(C_arrays: ArrayLike) -> tuple[Array, Array]:
    """Get the largest (eigenvalue, eigenvector) for each pixel in a 3D stack.

    Returns real eigenvalues, assuming the arrays are Hermitian.

    Parameters
    ----------
    C_arrays : ArrayLike
        The stack of coherence matrices.
        Shape = (rows, cols, nslc, nslc)

    Returns
    -------
    eigenvalues : Array
        The largest eigenvalue for each pixel's matrix
        Shape = (rows, cols)
    eigenvectors : Array
        The normalized eigenvector corresponding to the largest eigenvalue
        Shape = (rows, cols, nslc)

    """
    eig_vals, eig_vecs = vmap(vmap(power_iteration))(C_arrays)
    return eig_vals.real, eig_vecs


@partial(jit, static_argnames=("tol",))
def power_iteration(
    A: jnp.array, tol: float = 1e-5, max_iters: int = 50
) -> tuple[jnp.array, jnp.array]:
    """Compute the dominant eigenpair of a matrix using power iteration.

    Parameters
    ----------
    A : jnp.array
        The input matrix.
    tol : float, optional
        The tolerance for convergence (default is 1e-3).
    max_iters : int, optional
        The maximum number of iterations (default is 50).

    Returns
    -------
    eigenvalue : jnp.array
        The dominant eigenvalue of the matrix.
    vk : jnp.array
        The corresponding eigenvector of the dominant eigenvalue.

    """
    n = A.shape[-1]
    vk = jnp.ones(n, dtype=A.dtype)
    vk = vk / jnp.linalg.norm(vk)

    def body_fun(val):
        vk, _, idx = val
        A_times_vk = A @ vk
        next_vk = A_times_vk / jnp.linalg.norm(A_times_vk)
        diff = jnp.linalg.norm(next_vk - vk)
        return next_vk, diff, idx + 1

    def cond_fun(val):
        _, diff, iters = val
        return jnp.logical_and(diff > tol, iters < max_iters)

    init_val = (vk, 1, 1)
    vk, _, end_iters = while_loop(cond_fun, body_fun, init_val)

    # vk is normalized to 1, so no need to divide by (vk.T @ vk)
    eigenvalue = vk.conj() @ A @ vk
    return eigenvalue, vk


@partial(jit, static_argnames=("mu", "tol", "max_iters"))
def inverse_iteration(
    A: ArrayLike,
    mu: float,
    tol: float = 1e-5,
    max_iters: int = 50,
) -> tuple[Array, Array]:
    """Compute the eigenvalue of the positive definite matrix `A` closest to `mu`.

    Inverse iteration (or inverse power iteration) is an iterative method used to
    find the eigenvalue of a matrix A that is closest to a given scalar mu.

    Parameters
    ----------
    A : ArrayLike
        Square matrix of which we seek an eigenvalue and eigenvector.
    mu : float
        The shift value, around which we seek the closest eigenvalue of A.
    tol : float, optional
        Tolerance for convergence of the method. The default is 1e-5.
    max_iters : int, optional
        Maximum number of iterations to perform. The default is 50.

    Returns
    -------
    eigenvalue : jnp.array
        The eigenvalue of the matrix closest to `mu`.
    vk : jnp.array
        The corresponding eigenvector.

    Notes
    -----
    This method may not converge with a poor guess of `mu`.
    The close the guess of `mu` is to your eigenvalue, the quicker the convergence.
    However, picking the exact eigenvalue may lead to numerical instability.

    References
    ----------
    [1] https://services.math.duke.edu/~jtwong/math361-2019/lectures/Lec10eigenvalues.pdf

    """
    n = A.shape[0]
    vk = jnp.ones(n, dtype=A.dtype)

    vk = vk / jnp.linalg.norm(vk)
    Id = jnp.eye(A.shape[0], dtype=A.dtype)
    # Prefactor A - mu I to quickly solve each iteration
    lu_and_pivots = jax.scipy.linalg.lu_factor(A - mu * Id)

    def body_fun(val):
        vk_cur, _, iters = val
        # Perform power iteration on (A - mu I)^{-1} ,
        # which is the same as solving (A - mu I)x = v_k
        vk_new = jax.scipy.linalg.lu_solve(lu_and_pivots, vk_cur)
        vk_new = vk_new / jnp.linalg.norm(vk_new)
        diff = jnp.linalg.norm(vk_new - vk_cur)
        return vk_new, diff, iters + 1

    def cond_fun(val):
        _, diff, iters = val
        return jnp.logical_and(diff > tol, iters < max_iters)

    init_val = (vk, 1.0, 1)
    vk_sol, _, end_iters = while_loop(cond_fun, body_fun, init_val)

    eigenvalue = vk_sol.conj() @ A @ vk_sol
    return eigenvalue, vk_sol



import shutil
from datetime import datetime, timezone
from pathlib import Path

import pydantic
import pytest
from make_netcdf import create_test_nc

from dolphin.workflows import UnwrapMethod, config


# Testing what the defaults look like for each class
def test_half_window_defaults():
    hw = config.HalfWindow()
    assert hw.x == 11
    assert hw.y == 5
    assert hw.model_dump() == {"x": 11, "y": 5}


def test_half_window_to_looks():
    hw = config.HalfWindow()
    row_looks, col_looks = (11, 23)
    assert (row_looks, col_looks) == hw.to_looks()
    assert hw == config.HalfWindow.from_looks(row_looks, col_looks)

    # Test half window
    hw = config.HalfWindow(x=2, y=4)
    row_looks, col_looks = (9, 5)
    assert (row_looks, col_looks) == hw.to_looks()
    assert hw == config.HalfWindow.from_looks(row_looks, col_looks)


def test_ps_options_defaults():
    # Change directory so the creation of the default files doesn't fail
    pso = config.PsOptions()
    assert pso.amp_dispersion_threshold == 0.25
    assert pso._directory == Path("PS")
    assert pso._output_file == Path("PS/ps_pixels.tif")
    assert pso._amp_dispersion_file == Path("PS/amp_dispersion.tif")
    assert pso._amp_mean_file == Path("PS/amp_mean.tif")


def test_phase_linking_options_defaults():
    opts = config.PhaseLinkingOptions()
    assert opts.ministack_size == 10
    assert opts.half_window == config.HalfWindow()
    assert opts._directory == Path("linked_phase")
    assert opts.use_evd is False


def test_phase_linking_options_bad_size():
    with pytest.raises(pydantic.ValidationError):
        config.PhaseLinkingOptions(ministack_size=0)
        config.PhaseLinkingOptions(ministack_size=-1)


def test_interferogram_network_defaults():
    opts = config.InterferogramNetwork()
    assert opts.reference_idx == 0
    assert opts.max_bandwidth is None
    assert opts.max_temporal_baseline is None


def test_interferogram_network_types():
    opts = config.InterferogramNetwork(max_bandwidth=2)
    assert opts.max_bandwidth == 2
    assert opts.max_temporal_baseline is None

    opts = config.InterferogramNetwork(max_bandwidth=1)
    assert opts.max_bandwidth == 1

    opts = config.InterferogramNetwork(max_temporal_baseline=30)
    assert opts.max_temporal_baseline == 30
    assert opts.max_bandwidth is None

    with pytest.raises(pydantic.ValidationError):
        config.InterferogramNetwork(max_bandwidth=0)
        config.InterferogramNetwork(max_temporal_baseline=0)
        config.InterferogramNetwork(max_bandwidth=-1)
        config.InterferogramNetwork(max_temporal_baseline="asdf")


def test_unwrap_options_defaults():
    opts = config.UnwrapOptions()
    assert opts.unwrap_method == UnwrapMethod.SNAPHU
    assert opts._directory == Path("unwrapped")
    sn_opts = opts.snaphu_options
    assert sn_opts.init_method == "mcf"
    assert sn_opts.ntiles == (1, 1)
    assert sn_opts.cost == "smooth"
    tophu_opts = opts.tophu_options
    assert tophu_opts.downsample_factor == (1, 1)
    assert tophu_opts.ntiles == (1, 1)
    assert tophu_opts.cost == "smooth"


def test_outputs_defaults():
    opts = config.OutputOptions()
    assert opts.output_resolution is None
    assert opts.strides == {"x": 1, "y": 1}
    assert opts.hdf5_creation_options == {
        "chunks": [128, 128],
        "compression": "gzip",
        "compression_opts": 4,
        "shuffle": True,
    }


def test_worker_settings_defaults():
    ws = config.WorkerSettings()
    assert ws.gpu_enabled is False
    assert ws.threads_per_worker == 1
    assert ws.block_shape == (512, 512)


@pytest.fixture()
def dir_with_1_slc(tmp_path, slc_file_list_nc):
    p = tmp_path / "slc"
    p.mkdir()

    fname = "slc_20220101.nc"
    shutil.copy(slc_file_list_nc[0], p / fname)
    with open(p / "slclist.txt", "w") as f:
        f.write(fname + "\n")
    return p


@pytest.fixture()
def dir_with_2_slcs(tmp_path, slc_file_list_nc):
    p = tmp_path / "slc"
    p.mkdir()
    shutil.copy(slc_file_list_nc[0], p / "slc_20220101.nc")
    shutil.copy(slc_file_list_nc[1], p / "slc_20220102.nc")
    with open(p / "slclist.txt", "w") as f:
        f.write("slc_20220101.nc\n")
        f.write("slc_20220102.nc\n")
    return p


def test_inputs_defaults(dir_with_1_slc):
    # make a dummy file
    opts = config.DisplacementWorkflow(
        cslc_file_list=dir_with_1_slc / "slclist.txt",
        input_options={"subdataset": "data"},
    )

    assert opts.cslc_file_list[0].parent == dir_with_1_slc.resolve()
    assert opts.input_options.cslc_date_fmt == "%Y%m%d"
    assert len(opts.cslc_file_list) == 1
    assert isinstance(opts.cslc_file_list[0], Path)

    assert opts.mask_file is None

    # check it's coerced to a list of Paths
    opts2 = config.DisplacementWorkflow(
        cslc_file_list=[str(opts.cslc_file_list[0])],
        input_options={"subdataset": "data"},
    )
    assert isinstance(opts2.cslc_file_list[0], Path)


def test_inputs_bad_filename(tmp_path):
    # make a dummy file
    bad_cslc_file = tmp_path / "nonexistent_slclist.txt"
    with pytest.raises(pydantic.ValidationError, match="does not exist"):
        config.DisplacementWorkflow(cslc_file_list=bad_cslc_file)


def test_input_find_slcs(slc_file_list_nc):
    cslc_dir = Path(slc_file_list_nc[0]).parent

    opts = config.DisplacementWorkflow(
        cslc_file_list=slc_file_list_nc, input_options={"subdataset": "data"}
    )
    assert opts.cslc_file_list == slc_file_list_nc
    dict1 = opts.model_dump()

    opts2 = config.DisplacementWorkflow(
        cslc_file_list=cslc_dir / "slclist.txt", input_options={"subdataset": "data"}
    )
    dict2 = opts2.model_dump()

    dict1.pop("creation_time_utc")
    dict2.pop("creation_time_utc")
    assert dict1 == dict2


def test_input_relative_paths(tmpdir, slc_file_list_nc):
    with tmpdir.as_cwd():
        newfile = Path(slc_file_list_nc[0].name)
        shutil.copy(slc_file_list_nc[0], newfile)
        opts = config.DisplacementWorkflow(
            cslc_file_list=[newfile],
            input_options={"subdataset": "data"},
            keep_paths_relative=True,
        )
        assert opts.cslc_file_list == [newfile]


def test_input_glob_pattern(slc_file_list_nc):
    cslc_dir = Path(slc_file_list_nc[0]).parent
    slc_glob = str(cslc_dir / "20*nc")

    opts = config.DisplacementWorkflow(
        cslc_file_list=slc_glob, input_options={"subdataset": "data"}
    )
    assert opts.cslc_file_list == slc_file_list_nc


def test_input_nc_missing_subdataset(slc_file_list_nc):
    cslc_dir = Path(slc_file_list_nc[0]).parent

    with pytest.raises(pydantic.ValidationError, match="Must provide subdataset name"):
        config.DisplacementWorkflow(cslc_file_list=cslc_dir / "slclist.txt")


def test_input_slc_date_fmt(dir_with_2_slcs):
    expected_slcs = [Path(str(p)) for p in sorted(dir_with_2_slcs.glob("*.nc"))]

    opts = config.DisplacementWorkflow(
        cslc_file_list=expected_slcs, input_options={"subdataset": "data"}
    )
    assert opts.cslc_file_list == expected_slcs

    # If files don't all match the date format, it should error
    bad_date_slc = dir_with_2_slcs / "notadate.nc"
    shutil.copy(expected_slcs[0], bad_date_slc)
    new_file_list = dir_with_2_slcs.glob("*.nc")
    with pytest.raises(pydantic.ValidationError):
        opts = config.DisplacementWorkflow(
            cslc_file_list=new_file_list, input_options={"subdataset": "data"}
        )

    slc_file1 = dir_with_2_slcs / "2022-01-01.nc"
    slc_file2 = dir_with_2_slcs / "2022-01-02.nc"
    shutil.copy(expected_slcs[0], slc_file1)
    shutil.copy(expected_slcs[1], slc_file2)
    new_file_list = dir_with_2_slcs.glob("2022-*.nc")

    opts = config.DisplacementWorkflow(
        cslc_file_list=new_file_list,
        input_options={"cslc_date_fmt": "%Y-%m-%d", "subdataset": "data"},
    )
    assert opts.cslc_file_list == [slc_file1, slc_file2]

    # Check that we can get slcs by passing empty string
    # Should match all created files
    all_file_list = list(dir_with_2_slcs.glob("*.nc"))
    opts = config.DisplacementWorkflow(
        cslc_file_list=all_file_list,
        input_options={"cslc_date_fmt": "", "subdataset": "data"},
    )
    assert set(opts.cslc_file_list) == set(all_file_list)


def test_input_date_sort(dir_with_2_slcs):
    file_list = [Path(str(p)) for p in sorted(dir_with_2_slcs.glob("*.nc"))]
    opts = config.DisplacementWorkflow(
        cslc_file_list=file_list, input_options={"subdataset": "data"}
    )
    assert opts.cslc_file_list == file_list

    opts = config.DisplacementWorkflow(
        cslc_file_list=reversed(file_list), input_options={"subdataset": "data"}
    )
    assert opts.cslc_file_list == file_list


def test_input_opera_cslc(tmp_path, slc_stack):
    """Check that we recognize the OPERA filename format."""
    # Make a file with the OPERA name like OPERA_BURST_RE
    # r"t(?P<track>\d{3})_(?P<burst_id>\d{6})_(?P<subswath>iw[1-3])"
    start_date = 20220101
    d = tmp_path / "opera"
    d.mkdir()
    name_template = d / "t001_{burst_id}_iw1_{date}.nc"
    file_list = []
    for i in range(len(slc_stack)):
        burst_id = f"{i + 1:06d}"
        fname = str(name_template).format(burst_id=burst_id, date=str(start_date + i))
        create_test_nc(
            fname,
            epsg=32615,
            subdir="data",
            data_ds_name="VV",
            data=slc_stack[i],
        )
        file_list.append(Path(fname))

    opts = config.DisplacementWorkflow(
        cslc_file_list=file_list, input_options={"subdataset": "/data/VV"}
    )
    assert opts.cslc_file_list == file_list
    assert opts.input_options.subdataset == "/data/VV"


def test_input_cslc_empty():
    with pytest.raises(pydantic.ValidationError):
        config.DisplacementWorkflow(cslc_file_list=None)
        config.DisplacementWorkflow(cslc_file_list="")
        config.DisplacementWorkflow(cslc_file_list=[])


def test_config_displacement_workflow_defaults(dir_with_1_slc):
    c = config.DisplacementWorkflow(
        cslc_file_list=dir_with_1_slc / "slclist.txt",
        input_options={"subdataset": "data"},
    )
    # These should be the defaults
    assert c.output_options == config.OutputOptions()
    assert c.worker_settings == config.WorkerSettings()
    assert c.input_options == config.InputOptions(subdataset="data")
    assert c.work_directory == Path().resolve()

    # Check the defaults for the sub-configs, where the folders
    # should have been moved to the working directory
    assert c.ps_options._directory == Path("PS").resolve()
    assert c.ps_options._amp_mean_file == Path("PS/amp_mean.tif").resolve()

    p = Path("PS/amp_dispersion.tif")
    assert c.ps_options._amp_dispersion_file == p.resolve()

    assert c.phase_linking._directory == Path("linked_phase").resolve()

    assert c.interferogram_network._directory == Path("interferograms").resolve()
    assert c.interferogram_network.reference_idx == 0
    assert c.interferogram_network.max_bandwidth is None
    assert c.interferogram_network.max_temporal_baseline is None

    assert c.unwrap_options._directory == Path("unwrapped").resolve()

    now = datetime.now(timezone.utc)
    assert (now - c.creation_time_utc).seconds == 0


def test_config_create_dir_tree(tmpdir, slc_file_list_nc):
    fname0 = "slc_20220101.nc"
    shutil.copy(slc_file_list_nc[0], tmpdir / fname0)

    with tmpdir.as_cwd():
        c = config.DisplacementWorkflow(
            cslc_file_list=[fname0], input_options={"subdataset": "data"}
        )
        c.create_dir_tree()

        assert c.ps_options._directory.exists()
        assert c.interferogram_network._directory.exists()
        assert c.phase_linking._directory.exists()
        assert c.unwrap_options._directory.exists()

        # Check that the working directory is created
        assert Path().exists()

        for d in c._directory_list:
            assert d.exists()
            assert d.is_dir()


def test_config_roundtrip_dict(dir_with_1_slc):
    c = config.DisplacementWorkflow(
        cslc_file_list=dir_with_1_slc / "slclist.txt",
        input_options={"subdataset": "data"},
    )
    c_dict = c.model_dump()
    c2 = config.DisplacementWorkflow(**c_dict)
    assert c == c2


def test_config_roundtrip_json(dir_with_1_slc):
    c = config.DisplacementWorkflow(
        cslc_file_list=dir_with_1_slc / "slclist.txt",
        input_options={"subdataset": "data"},
    )
    c_json = c.model_dump_json()
    c2 = config.DisplacementWorkflow.model_validate_json(c_json)
    assert c == c2


def test_config_roundtrip_yaml(tmp_path, dir_with_1_slc):
    outfile = tmp_path / "config.yaml"
    c = config.DisplacementWorkflow(
        cslc_file_list=dir_with_1_slc / "slclist.txt",
        input_options={"subdataset": "data"},
    )
    c.to_yaml(outfile)
    c2 = config.DisplacementWorkflow.from_yaml(outfile)
    assert c == c2


def test_config_roundtrip_yaml_with_comments(tmp_path, dir_with_1_slc):
    outfile = tmp_path / "config.yaml"
    c = config.DisplacementWorkflow(
        cslc_file_list=dir_with_1_slc / "slclist.txt",
        input_options={"subdataset": "data"},
    )
    c.to_yaml(outfile, with_comments=True)
    c2 = config.DisplacementWorkflow.from_yaml(outfile)
    assert c == c2


def test_config_print_yaml_schema(tmp_path, dir_with_1_slc):
    outfile = tmp_path / "empty_schema.yaml"
    c = config.DisplacementWorkflow(
        cslc_file_list=dir_with_1_slc / "slclist.txt",
        input_options={"subdataset": "data"},
    )
    c.print_yaml_schema(outfile)


def test_config_ps_workflow_defaults(dir_with_1_slc):
    c = config.PsWorkflow(
        cslc_file_list=dir_with_1_slc / "slclist.txt",
        input_options={"subdataset": "data"},
    )

    assert c.input_options == config.InputOptions(subdataset="data")
    # Need to compare `model_fields` because the new instance of `PsOptions`
    assert c.output_options == config.OutputOptions()
    # has different private directories
    assert c.ps_options.model_fields == config.PsOptions().model_fields



from typing import Final

CONNCOMP_SUFFIX = ".unw.conncomp.tif"
CONNCOMP_SUFFIX_ZEROED = ".unw.conncomp.zeroed.tif"
UNW_SUFFIX = ".unw.tif"
UNW_SUFFIX_ZEROED = ".unw.zeroed.tif"

UINT16_MAX: Final = 65535

DEFAULT_CCL_NODATA: Final = UINT16_MAX
DEFAULT_UNW_NODATA: Final = 0



"""Estimate wrapped phase using batches of ministacks.

Initially based on [@Ansari2017SequentialEstimatorEfficient].
"""

from __future__ import annotations

import logging
from itertools import chain
from os import fspath
from pathlib import Path
from typing import Optional

from osgeo_utils import gdal_calc

from dolphin import io
from dolphin._types import Filename
from dolphin.io import VRTStack
from dolphin.stack import MiniStackPlanner

from .config import ShpMethod
from .single import run_wrapped_phase_single

logger = logging.getLogger(__name__)

__all__ = ["run_wrapped_phase_sequential"]


def run_wrapped_phase_sequential(
    *,
    slc_vrt_file: Filename,
    ministack_planner: MiniStackPlanner,
    ministack_size: int,
    half_window: dict,
    strides: Optional[dict] = None,
    mask_file: Optional[Filename] = None,
    ps_mask_file: Optional[Filename] = None,
    amp_mean_file: Optional[Filename] = None,
    amp_dispersion_file: Optional[Filename] = None,
    shp_method: ShpMethod = ShpMethod.NONE,
    shp_alpha: float = 0.05,
    shp_nslc: Optional[int] = None,
    use_evd: bool = False,
    beta: float = 0.00,
    block_shape: tuple[int, int] = (512, 512),
    baseline_lag: Optional[int] = None,
    **tqdm_kwargs,
) -> tuple[list[Path], list[Path], Path, Path]:
    """Estimate wrapped phase using batches of ministacks."""
    if strides is None:
        strides = {"x": 1, "y": 1}
    output_folder = ministack_planner.output_folder
    output_folder.mkdir(parents=True, exist_ok=True)
    ministacks = ministack_planner.plan(ministack_size)

    v_all = VRTStack.from_vrt_file(slc_vrt_file)
    logger.info(f"Full file range: {v_all.file_list[0]} to {v_all.file_list[-1]}")
    logger.info(f"Output folder: {output_folder}")
    logger.info(f"Number of ministacks of size {ministack_size}: {len(ministacks)}")

    if shp_nslc is None:
        shp_nslc = v_all.shape[0]

    # list where each item is [output_slc_files] from a ministack
    output_slc_files: list[list] = []
    # Each item is the temp_coh/shp_count file from a ministack
    temp_coh_files: list[Path] = []
    shp_count_files: list[Path] = []

    # function to check if a ministack has already been processed
    def already_processed(d: Path, search_ext: str = ".tif") -> bool:
        return d.exists() and len(list(d.glob(f"*{search_ext}"))) > 0

    # Solve each ministack using the current chunk (and the previous compressed SLCs)
    for ministack in ministacks:
        cur_output_folder = ministack.output_folder

        if already_processed(cur_output_folder):
            logger.info(f"Skipping {cur_output_folder} because it already exists.")
        else:
            cur_files = ministack.file_list
            start_end = ministack.real_slc_date_range_str
            logger.info(
                f"Processing {len(cur_files)} SLCs. Output folder: {cur_output_folder}"
            )
            cur_vrt = VRTStack(
                cur_files,
                outfile=output_folder / f"{start_end}.vrt",
                sort_files=False,
                subdataset=v_all.subdataset,
            )

            # Currently: we are always using the first SLC as the reference,
            # even if this is a compressed SLC.
            # Will need to change this if we want to accommodate the original
            # Sequential Estimator+Datum Adjustment method.
            reference_idx = 0
            run_wrapped_phase_single(
                slc_vrt_file=cur_vrt,
                ministack=ministack,
                output_folder=cur_output_folder,
                half_window=half_window,
                strides=strides,
                reference_idx=reference_idx,
                use_evd=use_evd,
                beta=beta,
                mask_file=mask_file,
                ps_mask_file=ps_mask_file,
                amp_mean_file=amp_mean_file,
                amp_dispersion_file=amp_dispersion_file,
                shp_method=shp_method,
                shp_alpha=shp_alpha,
                shp_nslc=shp_nslc,
                block_shape=block_shape,
                baseline_lag=baseline_lag,
                **tqdm_kwargs,
            )

        cur_output_files, cur_comp_slc_file, temp_coh_file, shp_count_file = (
            _get_outputs_from_folder(cur_output_folder)
        )
        output_slc_files.append(cur_output_files)
        temp_coh_files.append(temp_coh_file)
        shp_count_files.append(shp_count_file)

    ##############################################

    # Average the temporal coherence files in each ministack
    full_span = ministack_planner.real_slc_date_range_str
    output_temp_coh_file = output_folder / f"temporal_coherence_average_{full_span}.tif"
    output_shp_count_file = output_folder / f"shp_counts_average_{full_span}.tif"

    # we can pass the list of files to gdal_calc, which interprets it
    # as a multi-band file
    _average_rasters(temp_coh_files, output_temp_coh_file, "Float32")
    _average_rasters(shp_count_files, output_shp_count_file, "Int16")

    # Combine the separate SLC output lists into a single list
    all_slc_files = list(chain.from_iterable(output_slc_files))
    all_comp_slc_files = [ms.get_compressed_slc_info().path for ms in ministacks]

    out_pl_slcs = []
    for slc_fname in all_slc_files:
        slc_fname.rename(output_folder / slc_fname.name)
        out_pl_slcs.append(output_folder / slc_fname.name)

    comp_slc_outputs = []
    for p in all_comp_slc_files:
        p.rename(output_folder / p.name)
        comp_slc_outputs.append(output_folder / p.name)

    return out_pl_slcs, comp_slc_outputs, output_temp_coh_file, output_shp_count_file


def _get_outputs_from_folder(
    output_folder: Path,
) -> tuple[list[Path], Path, Path, Path]:
    cur_output_files = sorted(output_folder.glob("2*.slc.tif"))
    cur_comp_slc_file = next(output_folder.glob("compressed_*"))
    temp_coh_file = next(output_folder.glob("temporal_coherence_*"))
    shp_count_file = next(output_folder.glob("shp_counts_*"))
    # Currently ignoring to not stitch:
    # eigenvalues, estimator, avg_coh
    return cur_output_files, cur_comp_slc_file, temp_coh_file, shp_count_file


def _average_rasters(file_list: list[Path], outfile: Path, output_type: str):
    if len(file_list) == 1:
        file_list[0].rename(outfile)
        return

    logger.info(f"Averaging {len(file_list)} files into {outfile}")
    gdal_calc.Calc(
        NoDataValue=0,
        format="GTiff",
        outfile=fspath(outfile),
        type=output_type,
        quiet=True,
        overwrite=True,
        creation_options=io.DEFAULT_TIFF_OPTIONS,
        A=file_list,
        calc="numpy.nanmean(A, axis=0)",
    )



from __future__ import annotations

import datetime
import os
import warnings
from pathlib import Path

import numpy as np
import pytest
from make_netcdf import create_test_nc
from osgeo import gdal
from rasterio.errors import NotGeoreferencedWarning

# https://numba.readthedocs.io/en/stable/user/threading-layer.html#example-of-limiting-the-number-of-threads
if not os.environ.get("NUMBA_NUM_THREADS"):
    os.environ["NUMBA_NUM_THREADS"] = str(min(os.cpu_count(), 16))  # type: ignore[type-var]

from opera_utils import OPERA_DATASET_NAME

from dolphin.io import load_gdal, write_arr
from dolphin.phase_link import simulate

NUM_ACQ = 30


# Filter rasterio georeferencing warnings
@pytest.fixture(autouse=True)
def suppress_not_georeferenced_warning():
    """
    Pytest fixture to suppress NotGeoreferencedWarning in tests.

    This fixture automatically applies to all test functions in the module
    where it's defined, suppressing the specified warning.
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=NotGeoreferencedWarning)
        yield


# https://github.com/pytest-dev/pytest/issues/667#issuecomment-112206152
@pytest.fixture()
def random():
    np.random.seed(1234)
    simulate._seed(1234)


@pytest.fixture(scope="session")
def slc_stack():
    shape = (NUM_ACQ, 5, 10)
    sigma = 0.5
    data = np.random.normal(0, sigma, size=shape) + 1j * np.random.normal(
        0, sigma, size=shape
    )
    return data.astype(np.complex64)


@pytest.fixture()
def slc_date_list(slc_stack):
    start_date = datetime.datetime(2022, 1, 1)
    out = []
    dt = datetime.timedelta(days=1)
    for i in range(len(slc_stack)):
        out.append(start_date + i * dt)

    return out


@pytest.fixture()
def slc_file_list(tmp_path, slc_stack, slc_date_list):
    shape = slc_stack.shape
    # Write to a file
    driver = gdal.GetDriverByName("GTiff")
    d = tmp_path / "gtiff"
    d.mkdir()
    name_template = d / "{date}.slc.tif"

    file_list = []
    for cur_date, cur_slc in zip(slc_date_list, slc_stack):
        fname = str(name_template).format(date=cur_date.strftime("%Y%m%d"))
        file_list.append(Path(fname))
        ds = driver.Create(fname, shape[-1], shape[-2], 1, gdal.GDT_CFloat32)
        ds.GetRasterBand(1).WriteArray(cur_slc)
        ds = None

    # Write the list of SLC files to a text file
    with open(d / "slclist.txt", "w") as f:
        f.write("\n".join([str(f) for f in file_list]))
    return file_list


@pytest.fixture()
def slc_file_list_nc(tmp_path, slc_stack, slc_date_list):
    """Save the slc stack as a series of NetCDF files."""
    d = tmp_path / "32615"
    d.mkdir()
    name_template = d / "{date}.nc"
    file_list = []
    for cur_date, cur_slc in zip(slc_date_list, slc_stack):
        fname = str(name_template).format(date=cur_date.strftime("%Y%m%d"))
        create_test_nc(fname, epsg=32615, subdir="/", data=cur_slc)
        assert 'AUTHORITY["EPSG","32615"]]' in gdal.Open(fname).GetProjection()
        file_list.append(Path(fname))

    # Write the list of SLC files to a text file
    with open(d / "slclist.txt", "w") as f:
        f.write("\n".join([str(f) for f in file_list]))
    return file_list


@pytest.fixture()
def slc_file_list_nc_wgs84(tmp_path, slc_stack, slc_date_list):
    """Make one with lat/lon as the projection system."""

    d = tmp_path / "wgs84"
    d.mkdir()
    name_template = d / "{date}.nc"
    file_list = []
    for cur_date, cur_slc in zip(slc_date_list, slc_stack):
        fname = str(name_template).format(date=cur_date.strftime("%Y%m%d"))
        create_test_nc(fname, epsg=4326, subdir="/", data=cur_slc)
        assert 'AUTHORITY["EPSG","4326"]]' in gdal.Open(fname).GetProjection()
        file_list.append(Path(fname))

    with open(d / "slclist.txt", "w") as f:
        f.write("\n".join([str(f) for f in file_list]))
    return file_list


@pytest.fixture()
def slc_file_list_nc_with_sds(tmp_path, slc_stack, slc_date_list):
    """Save NetCDF files with multiple valid datsets."""
    d = tmp_path / "nc_with_sds"
    name_template = d / "{date}.nc"
    d.mkdir()
    file_list = []
    subdirs = ["/data", "/data2"]
    ds_name = "VV"
    for cur_date, cur_slc in zip(slc_date_list, slc_stack):
        fname = str(name_template).format(date=cur_date.strftime("%Y%m%d"))
        create_test_nc(
            fname, epsg=32615, subdir=subdirs, data_ds_name=ds_name, data=cur_slc
        )
        # just point to one of them
        file_list.append(f"NETCDF:{fname}:{subdirs[0]}/{ds_name}")

    # Write the list of SLC files to a text file
    with open(d / "slclist.txt", "w") as f:
        f.write("\n".join([str(f) for f in file_list]))
    return file_list


# Phase linking fixtures for one neighborhood tests


@pytest.fixture(scope="session")
def C_truth():
    C, truth = simulate.simulate_coh(
        num_acq=NUM_ACQ,
        Tau0=72,
        gamma_inf=0.95,
        gamma0=0.99,
        add_signal=True,
        signal_std=0.01,
    )
    return C, truth


@pytest.fixture()
def slc_samples(C_truth):
    C, _ = C_truth
    ns = 11 * 11
    return simulate.simulate_neighborhood_stack(C, ns)


# General utils on loading data/attributes


@pytest.fixture()
def raster_100_by_200(tmp_path):
    ysize, xsize = 100, 200
    # Create a test raster
    driver = gdal.GetDriverByName("ENVI")
    d = tmp_path / "raster_100_by_200"
    d.mkdir()
    filename = str(d / "test.bin")
    ds = driver.Create(filename, xsize, ysize, 1, gdal.GDT_CFloat32)
    data = np.random.randn(ysize, xsize) + 1j * np.random.randn(ysize, xsize)
    ds.WriteArray(data)
    ds.FlushCache()
    ds = None
    return filename


@pytest.fixture()
def tiled_raster_100_by_200(tmp_path):
    ysize, xsize = 100, 200
    tile_size = [32, 32]
    creation_options = [
        "COMPRESS=DEFLATE",
        "ZLEVEL=5",
        "TILED=YES",
        f"BLOCKXSIZE={tile_size[0]}",
        f"BLOCKYSIZE={tile_size[1]}",
    ]
    # Create a test raster
    driver = gdal.GetDriverByName("GTiff")
    d = tmp_path / "tiled"
    d.mkdir()
    filename = d / "20220101test.tif"
    ds = driver.Create(
        str(filename), xsize, ysize, 1, gdal.GDT_CFloat32, options=creation_options
    )
    data = np.random.randn(ysize, xsize) + 1j * np.random.randn(ysize, xsize)
    ds.WriteArray(data)
    ds.FlushCache()
    ds = None
    return filename


@pytest.fixture()
def tiled_file_list(tiled_raster_100_by_200):
    tmp_path = tiled_raster_100_by_200.parent
    outname2 = tmp_path / "20220102test.tif"
    gdal.Translate(str(outname2), str(tiled_raster_100_by_200))
    outname3 = tmp_path / "20220103test.tif"
    gdal.Translate(str(outname3), str(tiled_raster_100_by_200))
    return [tiled_raster_100_by_200, outname2, outname3]


@pytest.fixture()
def raster_10_by_20(tmp_path, tiled_raster_100_by_200):
    # Write a small image to a file
    d = tmp_path / "raster_10_by_20"
    d.mkdir()
    outname2 = d / "20220102small.tif"
    gdal.Translate(str(outname2), str(tiled_raster_100_by_200), height=10, width=20)
    return outname2


@pytest.fixture()
def raster_with_nan(tmp_path, tiled_raster_100_by_200):
    # Raster with one nan pixel
    start_arr = load_gdal(tiled_raster_100_by_200)
    nan_arr = start_arr.copy()
    nan_arr[0, 0] = np.nan
    output_name = tmp_path / "with_one_nan.tif"
    write_arr(
        arr=nan_arr,
        like_filename=tiled_raster_100_by_200,
        output_name=output_name,
        nodata=np.nan,
    )
    return output_name


@pytest.fixture()
def raster_with_nan_block(tmp_path, tiled_raster_100_by_200):
    # One full block of 32x32 is nan
    output_name = tmp_path / "with_nans.tif"
    nan_arr = load_gdal(tiled_raster_100_by_200)
    nan_arr[:32, :32] = np.nan
    write_arr(
        arr=nan_arr,
        like_filename=tiled_raster_100_by_200,
        output_name=output_name,
        nodata=np.nan,
    )
    return output_name


@pytest.fixture()
def raster_with_zero_block(tmp_path, tiled_raster_100_by_200):
    # One full block of 32x32 is nan
    output_name = tmp_path / "with_zeros.tif"
    out_arr = load_gdal(tiled_raster_100_by_200)
    out_arr[:] = 1.0

    out_arr[:32, :32] = 0
    write_arr(
        arr=out_arr,
        like_filename=tiled_raster_100_by_200,
        output_name=output_name,
        nodata=0,
    )
    return output_name


# For displacement/workflow tests


DX, DY = 5, -10
X0, Y0 = 246362.5, 3422995.0
GRID_MAPPING_DSET = "spatial_ref"


@pytest.fixture()
def opera_slc_files(tmp_path) -> list[Path]:
    """Save the slc stack as a series of NetCDF files."""
    start_date = 20220101
    shape = (4, 128, 128)
    slc_stack = (np.random.rand(*shape) + 1j * np.random.rand(*shape)).astype(
        np.complex64
    )

    d = tmp_path / "input_slcs"
    d.mkdir()
    file_list = []

    *group_parts, ds_name = OPERA_DATASET_NAME.split("/")
    group = "/".join(group_parts)
    for burst_id in ["t087_185683_iw2", "t087_185684_iw2"]:
        for i in range(len(slc_stack)):
            fname = d / f"{burst_id}_{start_date + i}_20231201.h5"
            if i == 0:
                datestr = f"{start_date}_{start_date + i}_{start_date + len(slc_stack)}"
                fname = d / f"COMPRESSED_{burst_id}_{datestr}.h5"
            yoff = Y0 + i * shape[0] / 2
            create_test_nc(
                fname,
                epsg=32615,
                data_ds_name=ds_name,
                # The "dummy" is so that two datasets are created in the file
                # otherwise GDAL doesn't respect the NETCDF:file:/path/to/nested/data
                subdir=[group, "dummy"],
                data=slc_stack[i],
                xoff=X0,
                yoff=yoff,
            )
            file_list.append(Path(fname))

    return file_list


@pytest.fixture()
def opera_slc_files_official(tmp_path) -> list[Path]:
    base = "OPERA_L2_CSLC-S1"
    ending = "20230101T100506Z_S1A_VV_v1.0"
    # expected = {
    # "t087_185678_iw2": [
    # Path(f"{base}_T087-185678-IW2_20180210T232711Z_{ending}"),
    start = datetime.datetime(2022, 1, 1, 1, 2, 3)
    dt = datetime.timedelta(days=1)
    shape = (4, 128, 128)
    phase_stack = np.empty((4, 128, 128), dtype="float32")
    ramp0 = np.ones((128, 1)) * np.arange(128).reshape(1, -1)
    ramp0 /= 128
    for idx in range(len(phase_stack)):
        cur_slope = 0.5 * 12 * idx
        phase_stack[idx] = cur_slope * ramp0
    # Add a little noise
    noise = 0.05 * (np.random.rand(*shape) + 1j * np.random.rand(*shape)).astype(
        np.complex64
    )
    slc_stack = np.exp(1j * phase_stack) + noise

    d = tmp_path / "input_slcs"
    d.mkdir()
    file_list = []

    *group_parts, ds_name = OPERA_DATASET_NAME.split("/")
    group = "/".join(group_parts)
    for burst_id in ["T087-185683-IW2", "T087-185684-IW2"]:
        for i in range(len(slc_stack)):
            date_str = (start + i * dt).strftime("%Y%m%dT%H%M%S")
            fname = d / f"{base}_{burst_id}_{date_str}_{ending}.h5"
            yoff = Y0 + i * shape[0] / 2
            create_test_nc(
                fname,
                epsg=32615,
                data_ds_name=ds_name,
                # The "dummy" is so that two datasets are created in the file
                # otherwise GDAL doesn't respect the NETCDF:file:/path/to/nested/data
                subdir=[group, "corrections"],
                data=slc_stack[i],
                yoff=yoff,
                xoff=X0,
                dx=DX,
                dy=DY,
            )
            file_list.append(Path(fname))

    return file_list


@pytest.fixture()
def opera_static_files_official(tmp_path) -> list[Path]:
    base = "OPERA_L2_CSLC-S1-STATIC"
    ending = "20140403_S1A_v1.0.h5"

    d = tmp_path / "input_static_layers"
    d.mkdir()
    file_list = []

    for burst_id in ["T087-185683-IW2", "T087-185684-IW2"]:
        fname = d / f"{base}_{burst_id}_{ending}"
        create_static_layer_h5(fname)
        file_list.append(Path(fname))

    return file_list


def create_static_layer_h5(filename):
    import h5py
    from pyproj import CRS

    with h5py.File(filename, "w") as f:
        grp = f.create_group("/data")
        shape = (4965, 19878)  # Adjust shape as needed

        # Create datasets with specified values
        dsets = []
        dsets.append(grp.create_dataset("layover_shadow_mask", shape, dtype=np.uint16))
        dsets.append(
            grp.create_dataset(
                "los_east", shape, dtype=np.float32, data=0.7 * np.ones(shape)
            )
        )
        dsets.append(
            grp.create_dataset(
                "los_north", shape, dtype=np.float32, data=-0.11 * np.ones(shape)
            )
        )
        # Make the "grid_mapping" dataset
        # https://github.com/corteva/rioxarray/blob/21284f67db536d9c104aa872ab0bbc261259e59e/rioxarray/rioxarray.py#L34
        # grid_ds = grp.create_dataset(GRID_MAPPING_DSET, data=32615)
        grid_ds = f.create_dataset(GRID_MAPPING_DSET, data=32615)
        grid_ds.attrs.update(CRS.from_epsg(32615).to_cf())
        crs_wkt = CRS.from_epsg(32615).to_wkt()
        crs_attrs = {"crs_wkt": crs_wkt, "spatial_ref": crs_wkt}
        grid_ds.attrs.update(crs_attrs)

        xs = np.arange(X0, X0 + shape[1] * DX, DX)
        ys = np.arange(Y0, Y0 + shape[0] * DY, DY)
        x_ds = grp.create_dataset(
            "x_coordinates", (shape[1],), dtype=np.float32, data=xs
        )
        y_ds = grp.create_dataset(
            "y_coordinates", (shape[0],), dtype=np.float32, data=ys
        )
        x_ds.make_scale()
        y_ds.make_scale()
        for ds in dsets:
            ds.attrs.update({"grid_mapping": GRID_MAPPING_DSET})
            ds.dims[1].attach_scale(x_ds)
            ds.dims[0].attach_scale(y_ds)


@pytest.fixture()
def weather_model_files():
    data_dir = Path("tests/data")
    with open(data_dir / "weather_model_files.txt") as f:
        return f.readlines()


@pytest.fixture()
def tec_files():
    data_dir = Path("tests/data")
    with open(data_dir / "tec_files.txt") as f:
        return f.readlines()


@pytest.fixture()
def dem_file(tmp_path, slc_stack):
    fname = tmp_path / "dem.tif"
    shape = slc_stack.shape
    dem = np.random.rand(shape[-2], shape[-1]) + 1000
    # Write to a file
    driver = gdal.GetDriverByName("GTiff")
    ds = driver.Create(str(fname), shape[-1], shape[-2], 1, gdal.GDT_Float32)
    ds.GetRasterBand(1).WriteArray(dem)
    ds = None
    return fname


# For unwrapping/overviews
@pytest.fixture()
def list_of_gtiff_ifgs(tmp_path, raster_100_by_200):
    x, y = np.meshgrid(np.arange(200), np.arange(100))
    # simulate a simple phase ramp
    phase = 0.003 * x + 0.002 * y
    ifg_list = []
    # simulate three interferograms
    constants = [1, 2, 3]
    for i in range(3):
        # Create a copy of the raster in the same directory
        f = tmp_path / f"ifg_{i}.int.tif"
        array = np.exp(1j * (phase + constants[i]))
        write_arr(
            arr=array,
            output_name=f,
            like_filename=raster_100_by_200,
            driver="GTiff",
        )
        ifg_list.append(f)

    return ifg_list


@pytest.fixture()
def list_of_envi_ifgs(tmp_path, raster_100_by_200):
    ifg_list = []
    for i in range(3):
        # Create a copy of the raster in the same directory
        f = tmp_path / f"ifg_{i}.int"
        ifg_list.append(f)
        write_arr(
            arr=np.ones((100, 200), dtype=np.complex64),
            output_name=f,
            like_filename=raster_100_by_200,
            driver="ENVI",
        )
        ifg_list.append(f)

    return ifg_list
